import datetime
from dataclasses import dataclass
from typing import Iterator
from typing import List
from typing import Optional
from urllib.parse import quote

import marshmallow
import requests
from marshmallow_dataclass import class_schema


@dataclass
class ReleaseAsset:
    name: str
    label: Optional[str]
    state: str  # enum: "uploaded", "open"
    download_count: int
    content_type: str
    size: int
    browser_download_url: str
    created_at: datetime.datetime
    updated_at: datetime.datetime

    class Meta:
        unknown = marshmallow.EXCLUDE


@dataclass
class Release:
    name: Optional[str]
    body: Optional[str]
    tag_name: str
    html_url: str
    draft: bool
    prerelease: bool
    created_at: datetime.datetime
    published_at: Optional[datetime.datetime]
    assets: List[ReleaseAsset]

    class Meta:
        unknown = marshmallow.EXCLUDE


ReleaseSchema = class_schema(Release)


def iter_releases(
    owner: str,
    repo: str,
    per_page: Optional[int] = None,
    *,
    github_token: Optional[str] = None,
) -> Iterator[Release]:
    """Fetch releases for a GitHub repository.

    This uses GitHub’s `REST API`_ to download release data, including asset metadata,
    for the specified repository.

    .. _REST API: https://docs.github.com/en/rest/releases/releases
    """
    session = requests.Session()
    session.headers["Accept"] = "application/vnd.github+json"
    session.headers["User-Agent"] = "barnhunt (https://github.com/barnhunt/barnhunt)"
    if github_token is not None:
        session.headers["Authorization"] = f"Bearer {github_token}"
    releases_schema = ReleaseSchema(many=True)

    url = f"https://api.github.com/repos/{quote(owner)}/{quote(repo)}/releases"
    if per_page is not None:
        url += f"?per_page={per_page:d}"

    while url is not None:
        response = session.get(url)
        response.raise_for_status()
        assert response.status_code == 200
        yield from releases_schema.load(response.json())

        next = response.links.get("next")
        url = next["url"] if next else None



import sys
from typing import Iterable

import pytest

from barnhunt.pager import Command
from barnhunt.pager import get_pager
from barnhunt.pager import Grouper
from barnhunt.pager import Pager
from barnhunt.pager import TTYPager


@pytest.fixture
def tty_output(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(sys.stdin, "isatty", lambda: True)


@pytest.mark.usefixtures("tty_output")
def test_get_pager_returns_ttypager() -> None:
    pager = get_pager(42)
    assert isinstance(pager, TTYPager)
    assert pager.group_size == 42


@pytest.mark.usefixtures("tty_output")
def test_get_pager_returns_grouper(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("sys.stdin.isatty", lambda: False)
    pager = get_pager(43)
    assert isinstance(pager, Grouper)
    assert pager.group_size == 43


def test_Grouper(capsys: pytest.CaptureFixture[str]) -> None:
    pager = Grouper(3)
    pager([f"{n:d}" for n in range(5)])
    output = capsys.readouterr().out
    assert output.split("\n") == ["0", "1", "2", "", "3", "4", "", ""]


class TestTTYPager:
    @pytest.fixture
    def pager(self) -> Pager:
        return TTYPager(2)

    @pytest.fixture(autouse=True)
    def patch_getchar(
        self, keys: Iterable[str], monkeypatch: pytest.MonkeyPatch
    ) -> None:
        keys_iter = iter(keys)

        def getchar() -> str:
            return next(keys_iter)

        monkeypatch.setattr("click.getchar", getchar)

    @pytest.mark.parametrize(
        "keys",
        [
            ("v", "q"),
        ],
    )
    def test(self, pager: Pager, capsys: pytest.CaptureFixture[str]) -> None:
        lines = ["one", "two", "three"]
        pager(lines)
        output = capsys.readouterr().out
        for line in lines:
            assert line in output
        assert "\a" not in output

    @pytest.mark.parametrize(
        "keys",
        [
            ("q",),
        ],
    )
    def test_first_page(self, pager: Pager, capsys: pytest.CaptureFixture[str]) -> None:
        lines = ["one", "two", "three"]
        pager(lines)
        output = capsys.readouterr().out
        for line in lines[:2]:
            assert line in output
        assert lines[2] not in output

    @pytest.mark.parametrize(
        "keys",
        [
            ("k", "q"),
        ],
    )
    def test_beep(self, pager: Pager, capsys: pytest.CaptureFixture[str]) -> None:
        lines = ["one", "two", "three"]
        pager(lines)
        output = capsys.readouterr().out
        assert "\a" in output

    @pytest.mark.parametrize(
        "keys, command",
        [
            (("v",), Command.PAGE_DOWN),
            (("\x1b", "v"), Command.PAGE_UP),
        ],
    )
    def test_get_cmd(self, pager: TTYPager, command: Command) -> None:
        assert pager._get_cmd() == command

    @pytest.mark.parametrize(
        "keys, command",
        [
            (("x", "q"), Command.QUIT),
            (("\x1b", "x", "r"), Command.REDRAW),
        ],
    )
    def test_get_cmd_unrecognized_key(
        self, pager: TTYPager, command: Command, capsys: pytest.CaptureFixture[str]
    ) -> None:
        assert pager._get_cmd() == command
        assert capsys.readouterr().out == "\a"


@pytest.mark.parametrize(
    "key, command",
    [
        (" ", Command.PAGE_DOWN),
        ("q", Command.QUIT),
    ],
)
def test_Command_lookup(key: str, command: Command) -> None:
    assert Command.lookup(key) == command


def test_Command_lookup_raises_KeyError() -> None:
    with pytest.raises(KeyError):
        Command.lookup("x")



import logging

import pytest

from barnhunt.inkscape.css import InlineCSS


class TestInlineCSS:
    def test_setitem(self) -> None:
        css = InlineCSS()
        css["display"] = "none"
        assert css.serialize() == "display:none;"

        css = InlineCSS("display: inline; Display: block")
        css["display"] = "none"
        assert css.serialize() == "display:none;"

        css = InlineCSS("text-align:center")
        css["display"] = "none"
        assert css.serialize() == "text-align:center;display:none;"

    def test_getitem(self) -> None:
        css = InlineCSS("display: inline; text-align:center; Display:block")
        assert css["display"] == "block"
        with pytest.raises(KeyError):
            css["missing"]

    def test_delitem(self) -> None:
        css = InlineCSS("display: inline; text-align:center; Display: block")
        del css["display"]
        assert css.serialize() == "text-align:center;"

    def test_iter(self) -> None:
        css = InlineCSS("DISPLAY: inline; text-align:center; Display:block")
        assert list(css) == ["DISPLAY", "text-align"]

    def test_len(self) -> None:
        css = InlineCSS("DISPLAY: inline; text-align:center; Display:block")
        assert len(css) == 2

    def test_repr(self) -> None:
        css = InlineCSS()
        assert repr(css) == "<InlineCSS []>"

    def test_parse_error(self, caplog: pytest.LogCaptureFixture) -> None:
        css = InlineCSS("display:inline; mistake;text-align:center")
        warnings = [r for r in caplog.records if r.levelno >= logging.WARNING]
        assert len(warnings) == 1
        assert "mistake" in warnings[0].getMessage()
        assert css.serialize() == "display:inline;text-align:center;"

    def test_warns_on_at_rule(self, caplog: pytest.LogCaptureFixture) -> None:
        style = "display:inline;@display print {display:block}"
        css = InlineCSS(style)
        warnings = [r for r in caplog.records if r.levelno >= logging.WARNING]
        assert len(warnings) == 1
        assert "@ rules" in warnings[0].getMessage()
        assert css.serialize() == style

    def test_str(self) -> None:
        css = InlineCSS("x: fü")
        assert str(css) == "x: fü;"



import cmd
import shlex


class DummyShellmodeInkscape(cmd.Cmd):
    intro = "DummyShellmodeInkscape"
    prompt = ">"

    def precmd(self, line: str) -> str:
        return " ".join(shlex.split(line))

    def do_true(self, line: str) -> bool:
        return False

    def do_echo(self, line: str) -> bool:
        print(f"warning: {line}")
        return False

    def do_quit(self, line: str) -> bool:
        return True

    def do_EOF(self, line: str) -> bool:
        return True


if __name__ == "__main__":
    DummyShellmodeInkscape().cmdloop()



from __future__ import annotations

import logging
import os
from itertools import chain
from itertools import count
from typing import BinaryIO
from typing import Collection
from typing import Iterable
from typing import Iterator
from typing import NamedTuple
from typing import Sequence

import jinja2
from lxml import etree

from .inkscape import svg
from .layerinfo import dwim_layer_info
from .layerinfo import LayerFlags
from .layerinfo import LayerInfoParser
from .templating import FileAdapter
from .templating import get_element_context
from .templating import is_string_literal
from .templating import render_template
from .templating import TemplateContext

log = logging.getLogger()


class Coursemap(NamedTuple):
    sort_order: int
    tree: svg.ElementTree
    context: TemplateContext
    basename: str
    description: str


class TemplateRenderer:
    def __init__(self, layer_info_parser: LayerInfoParser):
        self.parse_layer_info = layer_info_parser

    def __call__(
        self, tree: svg.ElementTree, context: TemplateContext
    ) -> svg.ElementTree:
        tree = svg.copy_etree(tree)
        for elem in tree.iter(svg.SVG_TSPAN_TAG):
            if elem.text and not is_string_literal(elem.text):
                if not self._is_hidden(elem):
                    local_context = self._get_local_context(elem, context)
                    try:
                        elem.text = render_template(elem.text, local_context)
                    except jinja2.TemplateError as ex:
                        log.error(f"Error expanding template in SVG file: {ex!s}")
        return tree

    def _is_hidden(self, elem: svg.Element) -> bool:
        return any(
            self._is_hidden_layer(ancestor) for ancestor in svg.ancestor_layers(elem)
        )

    def _is_hidden_layer(self, elem: svg.LayerElement) -> bool:
        assert svg.is_layer(elem)
        info = self.parse_layer_info(elem)
        return (info.flags & LayerFlags.HIDDEN) == LayerFlags.HIDDEN

    def _get_local_context(
        self, elem: svg.Element, parent_context: TemplateContext
    ) -> TemplateContext:
        context = parent_context.copy()
        context.update(get_element_context(elem, self.parse_layer_info))
        return context


class CourseMaps:
    default_context = {
        "overlays": (),
        "course": None,
        "overlay": None,
    }

    def __init__(
        self,
        layer_info_parser: LayerInfoParser,
        context: TemplateContext | None = None,
    ):
        self.parse_layer_info = layer_info_parser
        self.context = dict()
        if context is not None:
            self.context.update(context)
        self.context.update(self.default_context)

    def __call__(
        self, tree: svg.ElementTree
    ) -> Iterator[tuple[TemplateContext, svg.ElementTree]]:
        for path, hidden_layers in self._iter_overlays(tree.getroot()):
            base_context = self._get_context(path)
            for basename in self._get_output_basenames(path):
                if basename:
                    hide_layers = hidden_layers.union(
                        self._find_exclusions(basename, tree.getroot())
                    )
                else:
                    hide_layers = hidden_layers

                # Can not just omit layers which contain the source for visible "clones"
                hidden_clone_source_layers = svg.find_hidden_clone_source_layers(
                    tree, hidden_layers=hide_layers
                )
                # Ensure that all the hidden clone source layers have ids
                # so that we can identify them in the cloned tree
                ensure_id = svg.EnsureId(tree)
                hidden_layer_ids = set(map(ensure_id, hidden_clone_source_layers))

                # Copy tree, omitting hidden layers that do not contain source for
                # visible clones.
                pruned = svg.copy_etree(
                    tree, omit_elements=hide_layers - hidden_clone_source_layers
                )

                # Adjust visibility of remaining layers
                for layer, children in svg.walk_layers2(pruned.getroot()):
                    if layer.get("id") not in hidden_layer_ids:
                        svg.ensure_visible(layer)
                    else:
                        svg.set_hidden(layer)
                        children[:] = []

                if basename:
                    context = {**base_context, "output_basename": basename}
                else:
                    context = base_context
                yield context, pruned

    def _get_context(self, path: Sequence[svg.LayerElement]) -> TemplateContext:
        context = self.context.copy()
        if path:
            overlay = path[-1]
            local_context = get_element_context(overlay, self.parse_layer_info)
            local_context.pop("layer", None)
            context.update(local_context)
        return context

    def _get_output_basenames(
        self, path: Sequence[svg.LayerElement]
    ) -> Iterable[str] | tuple[None]:
        if path:
            overlay = path[-1]
            for layer in chain((overlay,), svg.ancestor_layers(overlay)):
                layer_info = self.parse_layer_info(layer)
                if layer_info.output_basenames:
                    return layer_info.output_basenames
        return (None,)

    def _iter_overlays(
        self, elem: svg.Element
    ) -> Iterator[tuple[list[svg.LayerElement], set[svg.LayerElement]]]:
        overlays, cruft = self._find_overlays(elem)

        if len(overlays) == 0:
            yield [], cruft
            return

        for overlay in overlays:
            other_overlays = set(overlays).difference([overlay])
            for path, hidden in self._iter_overlays(overlay):
                yield [overlay] + path, hidden | cruft | other_overlays

    def _find_overlays(
        self, elem: svg.Element
    ) -> tuple[list[svg.LayerElement], set[svg.LayerElement]]:
        overlays = []
        cruft = set()
        for node, children in svg.walk_layers2(elem):
            info = self.parse_layer_info(node)
            if info.flags & LayerFlags.HIDDEN:
                cruft.add(node)
                children[:] = []
            elif info.flags & LayerFlags.OVERLAY:
                overlays.append(node)
                children[:] = []
        return overlays, cruft

    def _find_exclusions(
        self, output_basename: str, elem: svg.Element
    ) -> Iterator[svg.LayerElement]:
        def get_includes(node: svg.LayerElement) -> Collection[str]:
            return self.parse_layer_info(node).include_in

        for node, children in svg.walk_layers2(elem):
            info = self.parse_layer_info(node)
            exclude = output_basename in info.exclude_from
            if not exclude and output_basename not in info.include_in:
                exclude = any(
                    output_basename in get_includes(sibling)
                    for sibling in svg.sibling_layers(node)
                )
            if exclude:
                yield node
                children[:] = []


BASENAME_TMPL = (
    "{% if output_basename -%}"
    "  {{ output_basename }}"
    "{% else -%}"
    '  {{ overlays|map("safepath")|join("/") }}'
    "{% endif %}"
)

DESCRIPTION_TMPL = (
    "{{ svgfile.name }}"
    "{% if overlays -%}"
    '  :{{ overlays|join("/") }}'
    "{% endif %}"
)


def _hash_dev_ino(svgfile: BinaryIO) -> int:
    st = os.fstat(svgfile.fileno())
    return hash((st.st_dev, st.st_ino))


def iter_coursemaps(svgfiles: Iterable[BinaryIO]) -> Iterator[Coursemap]:
    """Iterate over all coursemaps in svgfiles.

    Returns an iterable of (tree, context, basename, description) tuples,
    one for each map to be exported, for coursemaps in svgfiles.
    """
    counter = count()
    for svgfile in svgfiles:
        tree = etree.parse(svgfile)
        layer_info_parser = dwim_layer_info(tree)

        random_seed = svg.get_random_seed(tree)
        if random_seed is None:
            log.warning("%s: no random-seed is set in SVG file", svgfile.name)
            random_seed = _hash_dev_ino(svgfile)

        # Expand jinja templates in text within SVG file
        file_context = {
            "random_seed": random_seed,
            "svgfile": FileAdapter(svgfile),
        }

        render_templates = TemplateRenderer(layer_info_parser)
        # FIXME: tree = render_templates(tree, file_context)

        coursemapper = CourseMaps(layer_info_parser, file_context)
        for context, map_tree in coursemapper(tree):
            rendered_tree = render_templates(map_tree, context)
            yield Coursemap(
                sort_order=next(counter),
                tree=rendered_tree,
                context=context,
                basename=render_template(BASENAME_TMPL, context),
                description=render_template(DESCRIPTION_TMPL, context),
            )



import io
import subprocess
from typing import Any
from typing import AnyStr
from typing import Callable
from typing import Iterable
from typing import Mapping
from typing import overload
from typing import TextIO

from _typeshed import StrOrBytesPath

from .spawnbase import _BufferType
from .spawnbase import SpawnBase

class PopenSpawn(SpawnBase[AnyStr, _BufferType]):
    proc: subprocess.Popen[bytes]
    pid: int
    terminated: bool

    @overload
    def __init__(
        self: "PopenSpawn[str, io.StringIO]",
        cmd: Iterable[str],
        timeout: float = ...,
        maxread: int = ...,
        searchwindowsize: int | None = ...,
        logfile: TextIO | None = ...,
        cwd: StrOrBytesPath | None = ...,
        env: Mapping[str, str] | None = ...,
        encoding: str = ...,
        codec_errors: str = ...,
        preexec_fn: Callable[[], Any] | None = ...,
    ) -> None: ...
    @overload
    def __init__(
        self: "PopenSpawn[bytes, io.BytesIO]",
        cmd: Iterable[str],
        timeout: float = ...,
        maxread: int = ...,
        searchwindowsize: int | None = ...,
        logfile: TextIO | None = ...,
        cwd: StrOrBytesPath | None = ...,
        env: Mapping[str, str] | None = ...,
        encoding: None = ...,
        codec_errors: str | None = ...,
        preexec_fn: Callable[[], Any] | None = ...,
    ) -> None: ...
    def read_nonblocking(self, size: int, timeout: float | None) -> AnyStr: ...
    def write(self, s: AnyStr) -> None: ...
    def writelines(self, sequence: Iterable[AnyStr]) -> None: ...
    def send(self, s: AnyStr) -> int: ...
    def sendline(self, s: AnyStr = ...) -> int: ...
    def wait(self) -> int: ...
    def kill(self, sig: int) -> None: ...
    def sendeof(self) -> None: ...



from __future__ import annotations

import os
import shutil
from pathlib import Path

import pytest

from barnhunt.cli import default_inkscape_command


@pytest.fixture
def tests_dir() -> Path:
    return Path(__file__).parent


@pytest.fixture
def test1_pdf(tests_dir: Path) -> Path:
    return tests_dir / "test1.pdf"


@pytest.fixture
def test2_pdf(tests_dir: Path) -> Path:
    return tests_dir / "test2.pdf"


@pytest.fixture
def drawing_svg(tests_dir: Path) -> Path:
    return tests_dir / "drawing.svg"


@pytest.fixture
def tmp_drawing_svg(tmp_path: Path, drawing_svg: Path) -> Path:
    tmp_drawing_svg = tmp_path / drawing_svg.name
    shutil.copyfile(drawing_svg, tmp_drawing_svg)
    return tmp_drawing_svg


_inkscape_executable = shutil.which(
    os.environ.get("INKSCAPE_COMMAND", default_inkscape_command())
)


@pytest.fixture
def inkscape_executable() -> str:
    """Path to inkscape executable."""
    if _inkscape_executable is None:
        pytest.skip("test requires inkscape")
    return _inkscape_executable



import json
from pathlib import Path
from typing import Any
from typing import Dict
from zipfile import ZipFile

import marshmallow
import pytest
from packaging.requirements import Requirement
from packaging.version import Version

from barnhunt.installer.metadata import InvalidDistribution
from barnhunt.installer.metadata import Metadata
from barnhunt.installer.metadata import metadata_from_distdir
from barnhunt.installer.metadata import metadata_from_distzip
from barnhunt.installer.metadata import metadata_from_json


def test_Metadata_display_name() -> None:
    md = Metadata("foo", Version("1.0"))
    assert md.display_name == "foo"
    md.display_name = "foo.bar"
    assert md.display_name == "foo.bar"


def test_Metadata_canonical_name() -> None:
    md = Metadata("foo.bar", Version("1.0"))
    assert md.canonical_name == "foo-bar"


def test_metadata_from_json_requires_dist() -> None:
    md = metadata_from_json(
        {
            "name": "name",
            "version": "1.0",
            "requires_dists": ["barnhunt>1.1", Requirement("lxml")],
        }
    )
    requires_dists = md.requires_dists
    assert isinstance(requires_dists, list)
    assert [req.name for req in requires_dists] == ["barnhunt", "lxml"]


def test_metadata_from_json_bad_requires_dist() -> None:
    with pytest.raises(marshmallow.ValidationError):
        metadata_from_json(
            {"name": "name", "version": "1.0", "requires_dists": ["bad requirement"]}
        )


def test_metadata_from_json_requires_python() -> None:
    md = metadata_from_json(
        {"name": "name", "version": "1.0", "requires_python": ">3.7"}
    )
    assert Version("3.8") in md.requires_python
    assert Version("3.7") not in md.requires_python


def test_metadata_from_json_provides_extras() -> None:
    md = metadata_from_json(
        {"name": "name", "version": "1.0", "provides_extras": ["foo_bar"]}
    )
    assert md.provides_extras == ["foo-bar"]


@pytest.fixture
def json_data() -> Dict[str, Any]:
    return {
        "name": "test.name",
        "version": "0.42",
    }


def test_metadata_from_zipdist(tmp_path: Path, json_data: Dict[str, Any]) -> None:
    zip_path = tmp_path / "test.zip"
    with ZipFile(zip_path, "w") as zf:
        zf.writestr("org.dairiki.foo_bar/METADATA.json", json.dumps(json_data))
    with ZipFile(zip_path) as zf:
        md = metadata_from_distzip(zf)
    assert md.name == json_data["name"]


def test_metadata_from_zipdist_invalid(tmp_path: Path) -> None:
    zip_path = tmp_path / "test.zip"
    with ZipFile(zip_path, "w") as zf:
        zf.writestr("org.dairiki.foo_bar/foo.txt", "howdy")
    with ZipFile(zip_path) as zf:
        with pytest.raises(InvalidDistribution):
            metadata_from_distzip(zf)


def test_metadata_from_distdir(tmp_path: Path, json_data: Dict[str, Any]) -> None:
    dist_path = tmp_path / "test.dist"
    dist_path.mkdir()
    Path(dist_path, "METADATA.json").write_text(json.dumps(json_data))

    md = metadata_from_distdir(dist_path)
    assert md.name == json_data["name"]


def test_metadata_from_distdir_invalid(tmp_path: Path) -> None:
    Path(tmp_path, "test.empty").mkdir()
    Path(tmp_path, "test.badmeta").mkdir()
    Path(tmp_path, "test.badmeta", "METADATA.json").write_text("burp")
    Path(tmp_path, "test.notadir").touch()

    for dist_path in tmp_path.iterdir():
        with pytest.raises(InvalidDistribution):
            metadata_from_distdir(dist_path)



from __future__ import annotations

import enum
import logging
import re
from typing import Callable
from typing import Collection
from typing import NamedTuple
from typing import Sequence

from .inkscape import svg

log = logging.getLogger()


class LayerFlags(enum.Flag):
    HIDDEN = enum.auto()
    OVERLAY = enum.auto()

    @property
    def flag_char(self) -> str:
        assert self.name
        return self.name[0].lower()

    def __str__(self) -> str:
        return "".join(flag.flag_char for flag in self.__class__ if self & flag)

    @classmethod
    def parse(cls, s: str) -> LayerFlags:
        value = cls(0)
        lookup = {flag.flag_char: flag for flag in cls}
        for c in set(s):
            flag = lookup.get(c)
            if flag is not None:
                value |= flag
            else:
                log.warning("unknown character '%s' in flags %r", c, s)
        return value


class LayerInfo(NamedTuple):
    flags: LayerFlags
    label: str
    output_basenames: Sequence[str] = ()
    exclude_from: Collection[str] = ()
    include_in: Collection[str] = ()


def parse_flagged_layer_info(elem: svg.LayerElement) -> LayerInfo:
    label = svg.layer_label(elem)

    exclude_from = set()
    include_in = set()
    m = re.match(
        r"\["
        r"  (?P<flags>\w*)"
        r"  (?:\|*(?P<output_basenames>(?:,*\w[-\w\d]*)*))?"
        r"  (?P<exclusions>(?:,*[!=]\w[-\w\d]*)*)"
        r"\]\s*",
        label,
        re.X,
    )
    if m:
        flags = LayerFlags.parse(m.group("flags"))
        output_basenames = [
            basename for basename in m.group("output_basenames").split(",") if basename
        ]
        for exclusion in m.group("exclusions").split(","):
            if exclusion.startswith("!"):
                exclude_from.add(exclusion[1:])
            elif exclusion.startswith("="):
                include_in.add(exclusion[1:])
            else:
                assert exclusion == ""
        label = label[m.end() :]
    else:
        flags = LayerFlags(0)
        output_basenames = []
    return LayerInfo(flags, label, output_basenames, exclude_from, include_in)


def _is_top_level(layer: svg.LayerElement) -> bool:
    """Is layer a top-level layer."""
    return svg.parent_layer(layer) is None


def _label_matches(layer: svg.LayerElement, pat: str | re.Pattern[str]) -> bool:
    return re.search(pat, svg.layer_label(layer)) is not None


def obs_is_ring(layer: svg.LayerElement) -> bool:
    """Match the top-level "Ring" layer.

    This layer is always displayed.
    """
    assert svg.is_layer(layer)
    return _is_top_level(layer) and _label_matches(layer, r"(?i)\bring\b")


def obs_is_course(layer: svg.LayerElement) -> bool:
    """Match the top-level "Course" layers

    These layers are displayed, one per coursemap.
    """
    assert svg.is_layer(layer)
    return (
        _is_top_level(layer)
        and not obs_is_ring(layer)
        and _label_matches(
            layer, r"(?i)\b(instinct|novice|open|senior|master|crazy ?8s?|c8)\b"
        )
    )


def obs_is_cruft(layer: svg.LayerElement) -> bool:
    assert svg.is_layer(layer)
    return _is_top_level(layer) and not obs_is_course(layer) and not obs_is_ring(layer)


def obs_is_overlay(layer: svg.LayerElement) -> bool:
    assert svg.is_layer(layer)
    parent = svg.parent_layer(layer)
    if parent is None:
        return False
    return svg.layer_label(parent) == "Overlays" and any(
        obs_is_course(ancestor) for ancestor in svg.ancestor_layers(parent)
    )


def parse_obs_layer_info(elem: svg.LayerElement) -> LayerInfo:
    """This is the old "heuristic" scheme of determining which layers are
    overlays and and cruft.

    This scheme is obsolete. It has not been used for drawing course
    maps since 2018.

    """
    if obs_is_cruft(elem):
        flags = LayerFlags.HIDDEN
    elif obs_is_course(elem) or obs_is_overlay(elem):
        flags = LayerFlags.OVERLAY
    else:
        flags = LayerFlags(0)
    return LayerInfo(flags, svg.layer_label(elem))


LayerInfoParser = Callable[[svg.LayerElement], LayerInfo]


def dwim_layer_info(tree: svg.ElementTree) -> LayerInfoParser:
    """Deduce layout type."""

    def has_flags(elem: svg.LayerElement) -> bool:
        return parse_flagged_layer_info(elem).flags != LayerFlags(0)

    if not any(has_flags(elem) for elem in svg.walk_layers(tree.getroot())):
        # Old style with overlays and hidden layers identified by
        # matching layer labels against various regexps.
        return parse_obs_layer_info
    else:
        # New style with flags in layer labels.
        # E.g. "[o] Overlay Layer Label", "[h] Hidden Layer"
        return parse_flagged_layer_info



from __future__ import annotations

import os
import stat
from typing import Callable
from typing import Sequence

import jinja2
import pytest

from barnhunt.inkscape import svg
from barnhunt.templating import _hash_string
from barnhunt.templating import FileAdapter
from barnhunt.templating import get_element_context
from barnhunt.templating import is_string_literal
from barnhunt.templating import LayerAdapter
from barnhunt.templating import random_rats
from barnhunt.templating import render_template
from barnhunt.templating import safepath
from barnhunt.templating import TemplateContext
from testlib import svg_maker


class TestLayerAdapter:
    @pytest.fixture(scope="session")
    def top_layer(self) -> svg.LayerElement:
        layer = svg_maker.layer
        return layer(
            "Parent",
            id="toplevel",
            children=[
                layer(
                    "[o|somefile] Overlay",
                    id="overlay",
                    children=[layer("Sublayer", id="sublayer")],
                )
            ],
        )

    @pytest.fixture
    def toplevel(self, top_layer: svg.LayerElement) -> LayerAdapter:
        return LayerAdapter(top_layer)

    @pytest.fixture
    def overlay(self, top_layer: svg.LayerElement) -> LayerAdapter:
        assert svg.is_layer(top_layer[0])
        return LayerAdapter(top_layer[0])

    @pytest.fixture
    def sublayer(self, top_layer: svg.LayerElement) -> LayerAdapter:
        assert svg.is_layer(top_layer[0][0])
        return LayerAdapter(top_layer[0][0])

    def test_id(self, sublayer: LayerAdapter) -> None:
        assert sublayer.id == "sublayer"

    def test_label(self, sublayer: LayerAdapter) -> None:
        assert sublayer.label == "Sublayer"

    def test_output_basenames(
        self, overlay: LayerAdapter, sublayer: LayerAdapter
    ) -> None:
        assert overlay.output_basenames == ["somefile"]
        assert sublayer.output_basenames is None

    def test_flagged_label(self, overlay: LayerAdapter) -> None:
        assert overlay.label == "Overlay"

    def test_is_overlay(self, overlay: LayerAdapter, sublayer: LayerAdapter) -> None:
        assert overlay.is_overlay
        assert not sublayer.is_overlay

    def test_lineage(
        self, toplevel: LayerAdapter, overlay: LayerAdapter, sublayer: LayerAdapter
    ) -> None:
        assert list(sublayer.lineage) == [sublayer, overlay, toplevel]

    def test_overlay(
        self, toplevel: LayerAdapter, overlay: LayerAdapter, sublayer: LayerAdapter
    ) -> None:
        assert overlay.overlay == overlay
        assert sublayer.overlay == overlay
        assert toplevel.overlay is None

    def test_parent(self, sublayer: LayerAdapter) -> None:
        assert sublayer.parent is not None
        assert sublayer.parent.id == "overlay"
        assert sublayer.parent.parent is not None
        assert sublayer.parent.parent.id == "toplevel"
        assert sublayer.parent.parent.parent is None

    def test_hash(self, sublayer: LayerAdapter) -> None:
        assert hash(sublayer) == _hash_string("sublayer")

    def test_eq(self, toplevel: LayerAdapter, sublayer: LayerAdapter) -> None:
        assert sublayer != toplevel
        assert sublayer == LayerAdapter(sublayer.elem)

    def test_repr(self, sublayer: LayerAdapter) -> None:
        assert repr(sublayer) == "<LayerAdapter id=sublayer>"

    def test_str(self, sublayer: LayerAdapter) -> None:
        assert str(sublayer) == "Sublayer"


@pytest.mark.parametrize(
    "s, h",
    [
        ("", 628939552449298973),
        ("Fü", 1702320224449935351),
    ],
)
def test_hash_string(s: str, h: int) -> None:
    assert _hash_string(s) == h


class Test_get_element_context:
    @pytest.fixture(scope="session")
    def tree(self) -> svg.ElementTree:
        layer = svg_maker.layer
        text = svg_maker.text
        return svg_maker.tree(
            layer("[h] Junk", id="cruft"),
            layer("Outline", id="ring"),
            layer(
                "[o] T1 Master",
                id="t1master",
                children=[
                    layer(
                        "[o|blinds] Blind 1",
                        id="blind1",
                        children=[
                            text(
                                "{{ course.label }} - {{ overlay.label }}",
                                id="blind1_title",
                            )
                        ],
                    ),
                    layer("[o|build_notes] Build Notes", id="build"),
                ],
            ),
            layer(
                "[o] T1 Novice",
                id="t1novice",
                children=[text("{{ course.label }}", id="novice_title")],
            ),
        )

    GetContext = Callable[[str], TemplateContext]

    @pytest.fixture
    def get_context(self, tree: svg.ElementTree) -> GetContext:
        def get_context(id: str) -> TemplateContext:
            (elem,) = tree.xpath("//*[@id=$id]", namespaces=svg.NSMAP, id=id)
            return get_element_context(elem)

        return get_context

    def test_one_overlay(self, get_context: GetContext) -> None:
        novice = get_context("novice_title")
        assert novice["course"].label == "T1 Novice"
        assert novice.get("overlay") is None

    def test_two_overlays(self, get_context: GetContext) -> None:
        master = get_context("blind1_title")
        assert master["course"].label == "T1 Master"
        assert master["overlay"].label == "Blind 1"

    def test_no_layers(self, get_context: GetContext) -> None:
        context = get_context("root")
        assert context == {}

    @pytest.mark.parametrize("id_", ["blind1_title", "novice_title"])
    def test_output_basename(self, id_: str, get_context: GetContext) -> None:
        context = get_context(id_)
        assert "output_basenames" not in context


class TestFileAdapter:
    @pytest.fixture
    def srcfile(self) -> FileAdapter:
        return FileAdapter(open(__file__, "rb"))

    def test_name(self, srcfile: FileAdapter) -> None:
        assert srcfile.name == __file__

    def test_basename(self, srcfile: FileAdapter) -> None:
        assert srcfile.basename == os.path.basename(__file__)

    def test_stat(self, srcfile: FileAdapter) -> None:
        assert srcfile.stat == os.stat(__file__)
        assert stat.S_ISREG(srcfile.stat.st_mode)

    def test_hash(self, srcfile: FileAdapter) -> None:
        dev_ino = srcfile.stat.st_dev, srcfile.stat.st_ino
        assert hash(srcfile) == hash(dev_ino)

    def test_repr(self, srcfile: FileAdapter) -> None:
        assert repr(srcfile) == f"<FileAdapter {__file__!r}>"

    def test_str(self, srcfile: FileAdapter) -> None:
        assert str(srcfile) == __file__


def make_context(template_context: TemplateContext) -> jinja2.runtime.Context:
    return jinja2.runtime.Context(
        jinja2.Environment(),
        template_context,
        name=None,
        blocks={},
    )


class Test_random_rats:
    def check_plausibility(self, results: Sequence[int]) -> None:
        assert len(results) == 5
        assert all(1 <= r <= 5 for r in results)

    def test_basic(self) -> None:
        for n in range(20):
            rats = random_rats(make_context({"random_seed": n}))
            self.check_plausibility(rats)

    @pytest.mark.parametrize("varname", ["random_seed", "layer"])
    def test_results_change_if_seed_changes(self, varname: str) -> None:
        noseed = random_rats(make_context({}))
        assert any(noseed != random_rats(make_context({varname: n})) for n in range(20))
        assert noseed == random_rats(make_context({}))

    def test_skip(self) -> None:
        noskip = random_rats(make_context({}))
        rats = random_rats(make_context({}), skip=2)
        assert rats != noskip
        assert rats[:3] == noskip[2:]


@pytest.mark.parametrize(
    "pathcomp, expected",
    [
        ("a b", "a_b"),
        ("a/b", "a_b"),
    ],
)
def test_safepath(pathcomp: str, expected: str) -> None:
    assert safepath(pathcomp) == expected


def test_safepath_coerces_to_text() -> None:
    assert safepath(42) == "42"


class Test_render_template:
    @pytest.mark.parametrize(
        "tmpl, output",
        [
            ("foo", "foo"),
            ("{{ 'a b'|safepath }}", "a_b"),
            ("{{ rats(seed=0)|join(',') }}", "4,4,1,3,5"),
        ],
    )
    def test_render(self, tmpl: str, output: str) -> None:
        context: TemplateContext = {}
        assert render_template(tmpl, context) == output

    def test_strict_undefined(self) -> None:
        context: TemplateContext = {}
        with pytest.raises(jinja2.UndefinedError):
            assert render_template("{{x}}", context, strict_undefined=True)


def test_is_string_literal_true() -> None:
    assert is_string_literal("Some String")


@pytest.mark.parametrize(
    "tmpl",
    [
        "{{ somvar }}",
        "{% for x in [1] %}x{% endfor %}",
        "{% set x = 42 %}",
        "{# comment #}",
    ],
)
def test_is_string_literal_false(tmpl: str) -> None:
    assert not is_string_literal(tmpl)



from __future__ import annotations

import io
import locale
import logging
import os
import re
import shlex
import threading
import weakref
from dataclasses import dataclass
from subprocess import PIPE
from subprocess import Popen
from subprocess import run
from subprocess import STDOUT
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Sequence
from typing import TypeVar

from pexpect.popen_spawn import PopenSpawn

from .._compat import Protocol

log = logging.getLogger()


class InkscapeCommand(Protocol):
    @property
    def cli_args(self) -> Iterable[str]:
        """Commmand line args for Inkscape."""
        raise NotImplementedError()

    @property
    def shell_mode_cmdline(self) -> str:
        """Equivalent shell-mode command line"""
        raise NotImplementedError()


class ExportPdfCommand(InkscapeCommand):
    def __init__(self, svg_fn: str, output_fn: str, pdf_version: str | None = None):
        self.svg_fn = svg_fn
        self.output_fn = output_fn
        self.pdf_version = pdf_version

    @property
    def cli_args(self) -> list[str]:
        """Commmand line args for Inkscape."""
        return [self.svg_fn, *(f"--{param}" for param in map("=".join, self._params))]

    @property
    def _params(self) -> list[tuple[str, str] | tuple[str]]:
        raise NotImplementedError()


class ExportPdfCommand_1_0(ExportPdfCommand):
    @property
    def _params(self) -> list[tuple[str, str] | tuple[str]]:
        params: list[tuple[str, str] | tuple[str]] = [
            ("export-area-page",),
            ("export-type", "pdf"),
            ("export-dpi", "96"),  # for rasterization of filter effects
            ("export-filename", self.output_fn),
        ]
        if self.pdf_version is not None:
            params.insert(2, ("export-pdf-version", self.pdf_version))
        return params

    @property
    def shell_mode_cmdline(self) -> str:
        # Inkscape's current shell-mode parser appears to be quite rudimentary.
        # Split actions on \s*;\s*, then split max one parameter from each action
        # using \s*:\s*.  There appears to be no quoting mechanism.
        #
        # XXX: Not sure about encoding details.  Hope for the best for now.
        #
        # https://gitlab.com/inkscape/inkscape/-/blob/INKSCAPE_1_2_1/src/inkscape-application.cpp#L1119
        params: list[tuple[str, str] | tuple[str]] = [
            ("file-open", self.svg_fn),
            *self._params,
            ("export-do",),
        ]
        return "; ".join(map(":".join, params))


class ExportPdfCommand_0_9x(ExportPdfCommand):
    @property
    def _params(self) -> list[tuple[str, str] | tuple[str]]:
        params: list[tuple[str, str] | tuple[str]] = [
            ("export-area-page",),
            ("export-dpi", "96"),  # for rasterization of filter effects
            ("export-pdf", self.output_fn),
        ]
        if self.pdf_version is not None:
            params.insert(1, ("export-pdf-version", self.pdf_version))
        return params

    @property
    def shell_mode_cmdline(self) -> str:
        # Inkscape 0.9x's shell-mode parser uses Glib's g_shell_parse_argv
        # to parse command lines.  As such, I'm guessing it supports normal
        # shell quoting mechanisms.
        #
        # https://gitlab.com/inkscape/inkscape/-/blob/INKSCAPE_0_92_5/src/main.cpp#L1348
        return " ".join(map(shlex.quote, self.cli_args))


@dataclass
class InkscapeApi:
    export_pdf_command: Callable[[str, str, str | None], InkscapeCommand]
    cruft_patterns: tuple[str | re.Pattern[str], ...] = ()
    shell_mode_prompt: str = ">"
    shell_mode_cruft_patterns: tuple[str | re.Pattern[str], ...] = ()


# From RunApp in AppImage version
_appimage_cruft_patterns = (
    r"^Setting \$?(?:XDG_CONFIG_HOME|_INKSCAPE_GC)",
    r"^Run experimental bundle",  # (Inkscape 1.0.x only)
    r"^\s*$",
    r"^\(ld-linux.*?\):\s* Gtk-WARNING\b",
)
_gtk_cruft_patterns = (
    r"^(.*: )?Gtk-Message:",
    r"^(.*: )?Gtk-WARNING \*\*:",
)

INKSCAPE_APIS = {
    "1.0": InkscapeApi(
        export_pdf_command=ExportPdfCommand_1_0,
        cruft_patterns=(
            *_gtk_cruft_patterns,
            *_appimage_cruft_patterns,
        ),
        shell_mode_prompt="> ",
        shell_mode_cruft_patterns=(
            r"^Inkscape interactive shell mode\.",
            r"^\s*Input of the form:",
            r"^\s*action1:arg1;",
            r"^Only (?:verbs|actions) that don't require a desktop may be used\.",
            *_gtk_cruft_patterns,
            *_appimage_cruft_patterns,
        ),
    ),
    "0.9x": InkscapeApi(
        export_pdf_command=ExportPdfCommand_0_9x,
        cruft_patterns=(),
        shell_mode_prompt=">",
        shell_mode_cruft_patterns=(r"^Inkscape \d.* interactive shell mode",),
    ),
}


_Runner = TypeVar("_Runner", bound="Runner")


class Runner:
    def __init__(self, api: InkscapeApi, executable: str):
        self.api = api
        self.executable = executable

    def export_pdf(
        self, svg_fn: str, output_fn: str, pdf_version: str | None = None
    ) -> None:
        self.run(self.api.export_pdf_command(svg_fn, output_fn, pdf_version))

    def run(self, cmd: InkscapeCommand) -> None:
        raise NotImplementedError()

    def close(self) -> None:
        pass

    def __enter__(self: _Runner) -> _Runner:
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.close()


class CliRunner(Runner):
    """Run Inkscape in normal command-line mode.

    This will run Inkscape once for each operation.

    """

    def run(self, cmd: InkscapeCommand) -> None:
        proc = run(
            [self.executable, *cmd.cli_args],
            stdout=PIPE,
            stderr=STDOUT,
            text=True,
            check=True,
        )
        log_output(proc.stdout, self.api.cruft_patterns)


# Various environment variable seetings to attempt to disable GNU readline's
# horizontal-scroll-mode, which, if enabled, screws up pexpect's parsing of the command
# line when fed to `inkscape --shell`.
DISABLE_RL_HORIZ_SCROLL = {
    "INPUTRC": "/dev/null",
    "TERM": "dumb",
    "COLUMNS": "10000",
}


class ShellModeRunner(Runner):
    """Run Inkscape using it's --shell mode.

    This allows for running multiple commands with a single invocation
    of Inkscape.

    This runner is thread-safe.  If called from multiple threads, a
    separate inkscape process will be run for each thread.

    """

    def __init__(
        self,
        api: InkscapeApi,
        executable: str,
        timeout: float = 30,
        inkscape_args: Iterable[str] = ("--shell",),
    ):
        super().__init__(api, executable)
        self.timeout = timeout
        self.inkscape_args = inkscape_args
        self.prompt_re = r"(\A|\n)" + re.escape(api.shell_mode_prompt)
        self._threadlocal = threading.local()

    @property
    def child(self) -> PopenSpawn[str, io.StringIO]:
        threadlocal = self._threadlocal
        child: PopenSpawn[str, io.StringIO] | None = getattr(threadlocal, "child", None)
        if child is not None:
            return child

        log.debug("Starting shell-mode inkscape subprocess")
        encoding = locale.getpreferredencoding(False)
        child = threadlocal.child = PopenSpawn(
            [self.executable, *self.inkscape_args],
            encoding=encoding,
            searchwindowsize=len(self.api.shell_mode_prompt) + 1,
            timeout=self.timeout,
            env={**os.environ, **DISABLE_RL_HORIZ_SCROLL},
        )
        # Arrange to shutdown child when current thread is done
        # There are circular references within the PopenSpawn, so
        # it does not clean itself up very quickly, if at all.
        obj = threadlocal.sentinel = _Object()
        threadlocal.finalizer = weakref.finalize(obj, self._shutdown_child, child)
        self._wait_for_prompt(self.api.shell_mode_cruft_patterns)
        return child

    def run(self, cmd: InkscapeCommand) -> None:
        cmdline = cmd.shell_mode_cmdline
        log.debug("Sending to shell-mode inkscape: %r", cmdline)
        self.child.sendline(cmdline)
        self._wait_for_prompt([re.escape(cmdline), r"\A\s*\Z"])

    def close(self) -> None:
        finalizer = getattr(self._threadlocal, "finalizer", None)
        if callable(finalizer):
            finalizer()

    def _wait_for_prompt(self, cruft_patterns: Sequence[str | re.Pattern[str]]) -> None:
        """Wait for prompt."""
        self.child.expect(self.prompt_re)
        log_output(
            output=self.child.before,
            cruft_patterns=cruft_patterns,
            command="shell-mode Inkscape",
        )

    @staticmethod
    def _shutdown_child(child: PopenSpawn[str, io.StringIO]) -> None:
        child.sendeof()
        # FIXME: send kill()s, too, after a bit of a wait?

    @property
    def _proc(self) -> Popen[bytes] | None:
        """Access to the underlyting Popen object.

        This is for use by tests.
        """
        child: PopenSpawn[str, io.StringIO] | None = getattr(
            self._threadlocal, "child", None
        )
        if child is not None:
            return child.proc
        return None


class _Object:
    """Minimal object we can create a weakref to.

    (Weakref.finalizer doesn't work with plain object().)
    """


def log_output(
    output: str,
    cruft_patterns: Sequence[str | re.Pattern[str]] = (),
    command: str = "Inkscape",
) -> None:
    """Log output from command.

    If the output contains any lines which are not matched by one of
    the ``cruft_patterns``, a warning is logged, with the suspicious lines
    flagged.  Otherwise, the output is logged at DEBUG level.

    The ``command`` parameter should be set to a textual description of the
    command which generated the output.
    """

    def is_unexpected(line: str) -> bool:
        return all(re.search(pat, line) is None for pat in cruft_patterns)

    output_lines = output.splitlines()
    if any(is_unexpected(line) for line in output_lines):
        flag = "!! =>"
        log.warning(
            f"Unexpected output from {command} "
            f"(Unexpected lines are marked with {flag!r}):\n%s",
            "\n".join(
                f"{flag if is_unexpected(line) else '':<{len(flag)}s} {line}"
                for line in output_lines
            ),
        )
    elif output_lines:
        log.debug(
            f"Output from {command}:\n%s",
            "\n".join(f"  {line}" for line in output_lines),
        )


def inkscape_runner(
    shell_mode: bool = True,
    executable: str = "inkscape",
    old_inkscape: bool | None = None,
) -> Runner:
    if old_inkscape or (old_inkscape is None and dwim_old_inkscape(executable)):
        api = INKSCAPE_APIS["0.9x"]
    else:
        api = INKSCAPE_APIS["1.0"]

    runner: type[Runner] = ShellModeRunner if shell_mode else CliRunner
    return runner(api, executable)


def dwim_old_inkscape(executable: str) -> bool:
    """Determine whether we're running an old Inkscape."""
    proc = run([executable, "--version"], capture_output=True, text=True, check=True)
    m = re.search(r"(?m)^Inkscape (\d+)\.(\d+)\.(\d+) ", proc.stdout)
    if m is None:
        log.warning("Can not determine Inkscape version.")
        return False
    version = tuple(map(int, m.groups()))
    return version < (1, 0)



""" Helpers for dealing with inline CSS

"""
from __future__ import annotations

import logging
from typing import Iterable
from typing import Iterator
from typing import MutableMapping

from tinycss2 import ast
from tinycss2 import parse_component_value_list
from tinycss2 import parse_declaration_list
from tinycss2 import serialize

log = logging.getLogger()


def _warn_on_parse_errors(tokens: Iterable[ast.Node], input: str) -> None:
    for tok in tokens:
        if tok.type == "error":
            log.warning("Ignoring CSS parse error: %s, while parsing %r", tok, input)


def _parse_inline_css(input: str) -> list[ast.Node]:
    parsed = parse_declaration_list(input, skip_comments=True, skip_whitespace=True)
    _warn_on_parse_errors(parsed, input)
    if any(tok.type == "at-rule" for tok in parsed):
        log.warning("@ rules in CSS may not be handled correctly: %r", input)
    return parsed  # type: ignore[no-any-return]


class InlineCSS(MutableMapping[str, str]):
    def __init__(self, input: str | None = None):
        self._parsed = _parse_inline_css(input) if input else []

    def __getitem__(self, key: str) -> str:
        lower_key = ascii_lower(key)
        for tok in reversed(self._parsed):
            if tok.type == "declaration" and tok.lower_name == lower_key:
                # XXX: how to deal with tok.important?
                return serialize(tok.value)  # type: ignore[no-any-return]
        raise KeyError(key)

    def __setitem__(self, key: str, value: str) -> None:
        lower_key = ascii_lower(key)
        parsed_value = parse_component_value_list(value, skip_comments=True)
        _warn_on_parse_errors(parsed_value, value)

        self.__delitem__(key)
        tok = ast.Declaration(
            line=None,
            column=None,
            name=key,
            lower_name=lower_key,
            value=parsed_value,
            important=False,
        )
        self._parsed.append(tok)

    def __delitem__(self, key: str) -> None:
        lower_key = ascii_lower(key)
        self._parsed = [
            tok
            for tok in self._parsed
            if not (tok.type == "declaration" and tok.lower_name == lower_key)
        ]

    def __iter__(self) -> Iterator[str]:
        seen = set()
        for tok in self._parsed:
            if tok.type == "declaration":
                if tok.lower_name not in seen:
                    seen.add(tok.lower_name)
                    yield tok.name

    def __len__(self) -> int:
        return len(list(self.__iter__()))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._parsed!r}>"

    def serialize(self) -> str:
        return serialize(  # type: ignore[no-any-return]
            [tok for tok in self._parsed if tok.type != "error"]
        )

    __str__ = serialize


def ascii_lower(s: str) -> str:
    r"""Transform (only) ASCII letters to lower case.

    This is cribbed from webencodings.ascii_lower.
    """
    return s.encode("utf8").lower().decode("utf8")



from io import BytesIO

import pytest
from lxml import etree

from barnhunt.inkscape import svg
from testlib import get_by_id
from testlib import get_by_ids
from testlib import svg_maker

XML1 = b"""<root xmlns:foo="urn:example:foo">
  <a foo:attr="bar"><aa><aaa/></aa></a>
  <b/>
</root>"""


@pytest.fixture
def tree1() -> svg.ElementTree:
    return etree.parse(BytesIO(XML1))


SVG1 = b"""<svg xmlns:test="http://dairiki.org/testing"
                xmlns:bh="http://dairiki.org/barnhunt/inkscape-extensions"
                xmlns="http://www.w3.org/2000/svg"
                test:attr="test attr value"
                bh:random-seed="42"></svg>"""


@pytest.fixture
def svgtree1() -> svg.ElementTree:
    return etree.parse(BytesIO(SVG1))


@pytest.fixture
def coursemap1() -> svg.ElementTree:
    layer = svg_maker.layer
    text = svg_maker.text
    return svg_maker.tree(
        layer("Cruft"),
        layer("Ring", visible=False, children=[text("Ring", "ring_leaf")]),
        layer(
            "T1 Master",
            children=[
                layer(
                    "Overlays",
                    children=[
                        layer("Blind 1"),
                        layer(
                            "Build Notes",
                            children=[
                                text("Build Text"),
                            ],
                        ),
                    ],
                )
            ],
        ),
        layer("T1 Novice"),
    )


def test_is_layer(coursemap1: svg.ElementTree) -> None:
    matches = set(filter(svg.is_layer, coursemap1.iter()))
    assert {elem.get("id") for elem in matches} == {
        "t1novice",
        "t1master",
        "overlays",
        "blind1",
        "buildnotes",
        "ring",
        "cruft",
    }


def test_lineage(coursemap1: svg.ElementTree) -> None:
    overlays, t1master = get_by_ids(coursemap1, ("overlays", "t1master"))
    root = coursemap1.getroot()
    assert list(svg.lineage(overlays)) == [overlays, t1master, root]


def test_walk_layers(coursemap1: svg.ElementTree) -> None:
    layers = list(svg.walk_layers(coursemap1.getroot()))
    ids = [layer.get("id") for layer in layers]
    assert ids == [
        "t1novice",
        "t1master",
        "overlays",
        "buildnotes",
        "blind1",
        "ring",
        "cruft",
    ]


def test_walk_layers2(coursemap1: svg.ElementTree) -> None:
    layers = []
    for elem, children in svg.walk_layers2(coursemap1.getroot()):
        layers.append(elem)
        if elem.get("id") in ("t1master", "cruft"):
            children[:] = []
    assert [layer.get("id") for layer in layers] == [
        "t1novice",
        "t1master",
        "ring",
        "cruft",
    ]


def test_parent_layer(coursemap1: svg.ElementTree) -> None:
    overlays, t1master, ring, ring_leaf = get_by_ids(
        coursemap1, ("overlays", "t1master", "ring", "ring_leaf")
    )
    assert svg.parent_layer(t1master) is None
    assert svg.parent_layer(overlays) is t1master
    assert svg.parent_layer(ring_leaf) is ring


def test_ancestor_layers(coursemap1: svg.ElementTree) -> None:
    buildtext, buildnotes, overlays, t1master = get_by_ids(
        coursemap1, ("buildtext", "buildnotes", "overlays", "t1master")
    )
    assert list(svg.ancestor_layers(buildtext)) == [buildnotes, overlays, t1master]
    assert list(svg.ancestor_layers(buildnotes)) == [overlays, t1master]


def test_sibling_layers(coursemap1: svg.ElementTree) -> None:
    cruft, ring, t1master, t1novice = get_by_ids(
        coursemap1, ("cruft", "ring", "t1master", "t1novice")
    )
    assert list(svg.sibling_layers(ring)) == [cruft, t1master, t1novice]


def test_sibling_layers_no_parent(coursemap1: svg.ElementTree) -> None:
    root = coursemap1.getroot()
    assert list(svg.sibling_layers(root)) == []


@pytest.mark.parametrize(
    "style, expected",
    [
        ("display:none", "display:inline;"),
        ("display:none; text-align:center;", "text-align:center;display:inline;"),
        ("display:hidden", "display:hidden"),
        (None, None),
    ],
)
def test_ensure_visible(style: str, expected: str) -> None:
    attrib = {"style": style} if style is not None else {}
    elem = etree.Element("g", attrib=attrib)
    svg.ensure_visible(elem)
    assert elem.get("style") == expected


@pytest.mark.parametrize(
    "style, expected",
    [
        (None, "display:none;"),
        ("display:inline; text-align:center;", "text-align:center;display:none;"),
        ("display:none;text-align:center;", "display:none;text-align:center;"),
        ("display:hidden", "display:none;"),
    ],
)
def test_set_hidden(style: str, expected: str) -> None:
    attrib = {"style": style} if style is not None else {}
    elem = etree.Element("g", attrib=attrib)
    svg.set_hidden(elem)
    elem_style = elem.get("style")
    assert elem_style is not None
    assert set(elem_style.split(";")) == set(expected.split(";"))


def test_layer_label(coursemap1: svg.ElementTree) -> None:
    cruft = get_by_id(coursemap1, "cruft")
    assert svg.is_layer(cruft)
    assert svg.layer_label(cruft) == "Cruft"


def test_find_clones() -> None:
    ref = svg_maker.group(id="source")
    use = svg_maker.use(href="#source")
    tree = svg_maker.tree(svg_maker.layer("Layer 1", children=[ref, use]))
    assert list(svg.find_clones(tree)) == [(use, ref)]


def test_find_clones_ignores_clones_in_defs() -> None:
    ref = svg_maker.group(id="source")
    use = svg_maker.use(href="#source")
    tree = svg_maker.tree(
        svg_maker.defs(use),
        svg_maker.layer("Layer 1", children=[ref]),
    )
    assert list(svg.find_clones(tree)) == []


def test_find_clones_ignores_sources_in_defs() -> None:
    ref = svg_maker.group(id="source")
    use = svg_maker.use(href="#source")
    tree = svg_maker.tree(
        svg_maker.defs(ref),
        svg_maker.layer("Layer 1", children=[use]),
    )
    assert list(svg.find_clones(tree)) == []


def test_find_hidden_clones_source_layers() -> None:
    use = svg_maker.use(href="#source")
    ref = svg_maker.group(id="source")
    use2 = svg_maker.use(href="#source2")
    ref2 = svg_maker.group(id="source2")
    use3 = svg_maker.use(href="#sym")
    sym = svg_maker.group(id="sym")
    layer2 = svg_maker.layer("Layer 2", children=[ref])
    tree = svg_maker.tree(
        svg_maker.defs(sym),
        svg_maker.layer("Layer 1", children=[use, use2, use3]),
        layer2,
        svg_maker.layer("Layer 3", children=[ref2]),
    )
    assert svg.find_hidden_clone_source_layers(tree, {layer2}) == {layer2}


def test_find_hidden_clones_source_layers_skips_hidden_clones() -> None:
    use = svg_maker.use(href="#source")
    ref = svg_maker.group(id="source")
    layer1 = svg_maker.layer("Layer 1", children=[use])
    layer2 = svg_maker.layer("Layer 2", children=[ref])
    tree = svg_maker.tree(layer1, layer2)
    assert svg.find_hidden_clone_source_layers(tree, {layer1, layer2}) == set()


def test_find_hidden_clones_source_layers_follows_hidden_clones() -> None:
    use = svg_maker.use(href="#use2")
    use2 = svg_maker.use(href="#source", id="use2")
    ref = svg_maker.group(id="source")
    layer1 = svg_maker.layer("Layer 1", children=[use])
    layer2 = svg_maker.layer("Layer 2", children=[use2])
    layer3 = svg_maker.layer("Layer 3", children=[ref])
    tree = svg_maker.tree(layer1, layer2, layer3)
    hidden_clone_source_layers = svg.find_hidden_clone_source_layers(
        tree, {layer2, layer3}
    )
    assert hidden_clone_source_layers == {layer2, layer3}


class Test_copy_etree:
    def test_copy(self, tree1: svg.ElementTree) -> None:
        copy1 = svg.copy_etree(tree1)
        assert etree.tostring(copy1) == etree.tostring(tree1)
        assert copy1 is not tree1
        assert copy1.getroot() is not tree1.getroot()

    def test_update_nsmap(self, tree1: svg.ElementTree) -> None:
        copy1 = svg.copy_etree(tree1, update_nsmap={"bar": "urn:example:bar"})
        b = copy1.find("b")
        assert b is not None
        b.set("{urn:example:bar}y", "z")
        assert '<b bar:y="z"/>' in etree.tostring(copy1).decode("utf8")

    def test_omit_elements(self, tree1: svg.ElementTree) -> None:
        copy1 = svg.copy_etree(tree1, omit_elements=(tree1.find("aa")))
        assert copy1.find("aa") is None
        assert copy1.find("aaa") is None


def test_EnsureId_returns_existing_id() -> None:
    elem = svg_maker.group(id="existing-id")
    tree = svg_maker.tree(
        svg_maker.layer("Layer 1", children=[elem]),
    )
    ensure_id = svg.EnsureId(tree)
    assert ensure_id(elem) == "existing-id"


def test_EnsureId_generates_unique_id() -> None:
    elems = [svg_maker.group() for _ in range(100)]
    tree = svg_maker.tree(
        svg_maker.layer("Layer 1", children=elems),
    )
    ensure_id = svg.EnsureId(tree, mindigits=0, sparedigits=0)
    ids = [ensure_id(elem) for elem in elems]
    print(f"ids={ids!r}")
    assert len(set(ids)) == len(elems)


@pytest.mark.parametrize(
    "attr, expect",
    [
        ("attr", "test attr value"),
        ("unknown", "DEFVAL"),
    ],
)
def test_get_svg_attrib(svgtree1: svg.ElementTree, attr: str, expect: str) -> None:
    attr = etree.QName("http://dairiki.org/testing", attr).text
    assert svg.get_svg_attrib(svgtree1, attr, "DEFVAL") == expect


def test_get_svg_attrib_raises_value_error(tree1: svg.ElementTree) -> None:
    with pytest.raises(ValueError):
        svg.get_svg_attrib(tree1, "attr")


def test_set_svg_attrib(svgtree1: svg.ElementTree) -> None:
    attr = etree.QName("http://dairiki.org/testing", "newattr")
    svg.set_svg_attrib(svgtree1, attr, "new value")
    assert svgtree1.getroot().get(attr) == "new value"


def test_set_svg_attrib_raises_value_error(tree1: svg.ElementTree) -> None:
    with pytest.raises(ValueError):
        svg.set_svg_attrib(tree1, "attr", "value")


def test_get_random_seed(svgtree1: svg.ElementTree) -> None:
    assert svg.get_random_seed(svgtree1) == 42


def test_get_random_seed_default(svgtree1: svg.ElementTree) -> None:
    del svgtree1.getroot().attrib[svg.BH_RANDOM_SEED]
    assert svg.get_random_seed(svgtree1, 13) == 13


def test_get_random_seed_raise_value_error(svgtree1: svg.ElementTree) -> None:
    svgtree1.getroot().attrib[svg.BH_RANDOM_SEED] = "not an int"
    with pytest.raises(ValueError) as excinfo:
        svg.get_random_seed(svgtree1)
    assert "Expected integer" in str(excinfo.value)


def test_set_random_seed(svgtree1: svg.ElementTree) -> None:
    svg.set_random_seed(svgtree1, 43)
    assert svgtree1.getroot().attrib[svg.BH_RANDOM_SEED] == "43"


def test_set_random_seed_raises_value_error(svgtree1: svg.ElementTree) -> None:
    with pytest.raises(ValueError):
        svg.set_random_seed(svgtree1, "44")  # type: ignore[arg-type]



import logging
import os
import shutil
import sys
from pathlib import Path
from subprocess import run


log = logging.getLogger()


def get_user_data_directory(inkscape_command: str = "inkscape") -> Path:
    """Attempt to determine Inkscape’s user profile directory in various ways."""
    try:
        return _user_data_directory(inkscape_command)
    except Exception as exc:
        log.warning("failed to query inkscape for profile directory: %s", exc)
        return get_default_user_data_directory()


def _user_data_directory(inkscape_command: str = "inkscape") -> Path:
    """Run ``inkscape --user-data-directory`` to determine path to profile directory."""
    inkscape = shutil.which(inkscape_command)
    if inkscape is None:
        raise FileNotFoundError(f"command {inkscape_command!r} not found")
    cmd = (inkscape, "--user-data-directory")
    proc = run(cmd, capture_output=True, text=True, check=True)
    # Use final line: Inkscape 1.0 AppImage prints some initial cruft.
    _, _, user_data_dir = proc.stdout.strip().rpartition("\n")
    if not user_data_dir:
        raise ValueError(f"{' '.join(cmd)!r} did not produce any output")
    return Path(user_data_dir)


def get_default_user_data_directory() -> Path:
    """Attempt to determine Inkscape’s user profile directory
    from the environment.

    This attempts to duplicate the way that Inkscape itself deduces the
    user profile directory.
    """
    # Get the base configuration directory used by Inkscape.
    #
    # On Windows this appears to be %APPDATA% (typically
    # C:\Users\UserName\AppData\Roaming).
    #
    # On POSIX systems this is $XDG_CONFIG_HOME (default ~/.config).
    if sys.platform == "win32":
        config_home = Path(_get_appdata())
    elif "XDG_CONFIG_HOME" in os.environ:
        config_home = Path(os.environ["XDG_CONFIG_HOME"])
    else:
        config_home = Path.home() / ".config"

    return config_home / "inkscape"


def _get_appdata() -> str:
    """Get path to the Windows %APPDATA% directory."""
    from ctypes import c_int, create_unicode_buffer
    from ctypes import windll  # type: ignore[attr-defined]
    from ctypes.wintypes import DWORD, HANDLE, HWND, MAX_PATH

    CSIDL_APPDATA = 26

    path_buf = create_unicode_buffer(MAX_PATH)
    result = windll.shell32.SHGetFolderPathW(
        HWND(), c_int(CSIDL_APPDATA), HANDLE(), DWORD(), path_buf
    )
    if result != 0:
        raise RuntimeError(  # pragma: NO COVER
            f"ctypes call to SHGetFolderPathW failed ({result})"
        )
    return path_buf.value



""" Code for rendering Inkscape SVG to PDFS
"""
from __future__ import annotations

import re
import shutil
import time
from itertools import chain
from itertools import zip_longest
from pathlib import Path
from typing import BinaryIO
from typing import Iterable
from typing import Sequence

from pikepdf import ObjectStreamMode
from pikepdf import Pdf
from pikepdf import Rectangle

from ._compat import metadata


def concat_pdfs(in_fns: Sequence[str | Path], out_fn: str | Path) -> None:
    """Concatenate named PDF files to a single output file."""
    if len(in_fns) == 0:
        raise ValueError("No PDFs to concatenate")
    Path(out_fn).parent.mkdir(parents=True, exist_ok=True)
    if len(in_fns) == 1:
        shutil.copy(in_fns[0], out_fn)
    else:
        pdf = Pdf.new()
        for n, in_fn in enumerate(in_fns):
            src = Pdf.open(in_fn)
            if n == 0:
                update_metadata(pdf, src)
            pdf.pages.extend(src.pages)
        save_pdf(pdf, out_fn)


def two_up(
    infiles: Iterable[BinaryIO], outfile: BinaryIO, width: int = 612, height: int = 792
) -> None:
    """Generate 2-up version of PDF file(s)."""
    in_pdfs = [Pdf.open(fp) for fp in infiles]
    if len(in_pdfs) == 0:
        raise ValueError("No PDFs to concatenate")
    in_pages = list(chain.from_iterable(pdf.pages for pdf in in_pdfs))

    locations = [
        (Rectangle(0, height / 2, width, height), 270),
        (Rectangle(0, 0, width, height / 2), 90),
    ]
    n_out = (len(in_pages) + 1) // 2

    pdf = Pdf.new()
    update_metadata(pdf, in_pdfs[0])

    for pair in zip_longest(in_pages[:n_out], in_pages[n_out:]):
        out_page = pdf.add_blank_page(page_size=(width, height))
        for in_page, (box, rotation) in zip(pair, locations):
            if in_page is not None:
                in_page.rotate(rotation, relative=True)
                # XXX: should we pass push_stack=False to add_overlay?
                out_page.add_overlay(in_page, box)
    save_pdf(pdf, outfile)


def update_metadata(
    dst: Pdf, src: Pdf | None = None, now_: float | None = None
) -> None:
    """Update PDF metadata of ``dst``

    This sets the ``pdf:Producer`` to a string of the form "barnhunt <ourversion>",
    and ``xmp:MetadataDate`` to the current date/time.

    If ``src`` is given, additional metadata will be copied from that PDF.

    """
    with dst.open_metadata(set_pikepdf_as_editor=False) as meta:
        if src is not None:
            src_meta = src.open_metadata()
            if src_meta:
                meta.update(src_meta)
            else:
                meta.load_from_docinfo(src.docinfo)

        date = iso_date(now_)
        meta["pdf:Producer"] = f"barnhunt {metadata.version('barnhunt')}"
        meta["xmp:MetadataDate"] = date
        meta.setdefault("xmp:ModifyDate", date)


def iso_date(secs: float | None = None) -> str:
    """Format date and time to ISO-8601 format.

    Format is as described in https://www.w3.org/TR/NOTE-datetime

    Parameters:

    secs — Number of seconds since the epoch (e.g. as returned by
        ``time.time()``), or ``None`` for the current time.
    """
    now = time.strftime("%Y-%m-%dT%H:%M:%S%z", time.localtime(secs))
    # Add colon to zone offset
    return re.sub(r"(?<=[-+]\d\d)(?=\d\d\Z)", ":", now)


def save_pdf(pdf: Pdf, outfile: str | Path | BinaryIO) -> None:
    """Save pdf to file.

    This handles setting a bunch of pikepdf's (qpdf's) optimization options.
    (I'm not sure how much most of these help us.)
    """
    pdf.remove_unreferenced_resources()  # XXX: necessary, in our case?
    pdf.save(
        outfile,
        # This does reduce file size. Does it hurt us at all?
        object_stream_mode=ObjectStreamMode.generate,
        recompress_flate=True,  # Does this help us?
        linearize=True,  # Does this help us?
    )



from __future__ import annotations

import datetime
import io
import itertools
import json
import os
import re
import zipfile
from pathlib import Path
from pathlib import PurePosixPath
from typing import Iterable
from urllib.parse import urlunsplit

import pytest
from packaging.utils import canonicalize_name
from packaging.version import Version
from pytest_mock import MockerFixture

from barnhunt.installer import _copy_to_tmp
from barnhunt.installer import DownloadUrl
from barnhunt.installer import find_distributions
from barnhunt.installer import find_installed
from barnhunt.installer import github
from barnhunt.installer import InkexProject
from barnhunt.installer import InkexRequirement
from barnhunt.installer import Installer
from barnhunt.installer import NoSuchDistribution
from barnhunt.installer import open_zipfile

from ratelimit import mayberatelimited  # noreorder (test library)


def test_InkexRequirement_project() -> None:
    req = InkexRequirement("inkex_bh")
    assert req.project.install_dir == "extensions"
    assert req.project.gh_owner == "barnhunt"
    assert req.project.gh_repo == "inkex-bh"


@pytest.mark.parametrize(
    "requirement, message",
    [
        ("unknown", "unknown requirement"),
        ("inkex-bh[extra]", "extras not"),
        ("inkex-bh; python_version>'3.6'", "markers not"),
        ("unparseable requirement", "(?i)parse error|expected end"),
    ],
)
def test_InkexRequirement_project_raises_value_error(
    requirement: str, message: str
) -> None:
    with pytest.raises(ValueError) as exc_info:
        InkexRequirement(requirement)
    assert exc_info.match(message)


@pytest.fixture
def zip_data() -> bytes:
    with io.BytesIO() as fp:
        with zipfile.ZipFile(fp, "w") as zf:
            zf.writestr("test.txt", "howdy")
        return fp.getvalue()


def test_copy_to_tmp() -> None:
    orig = io.BytesIO(b"content")
    with _copy_to_tmp(orig) as fp:
        assert fp.read() == b"content"
        assert fp.read() == b""
        assert fp.seekable()
        fp.seek(0)
        assert fp.read() == b"content"


def test_open_zipfile(mocker: MockerFixture, zip_data: bytes) -> None:
    urlopen = mocker.patch("barnhunt.installer.urlopen")
    urlopen.return_value = io.BytesIO(zip_data)
    with open_zipfile(DownloadUrl("https://example.com/dummy.zip")) as zf:
        assert zf.namelist() == ["test.txt"]


@pytest.mark.requiresinternet
def test_functional_download_zipfile() -> None:
    url = DownloadUrl(
        "https://github.com/barnhunt/inkex-bh/releases/download/v1.0.0rc3/"
        "inkex_bh-1.0.0rc3.zip"
    )
    with open_zipfile(url) as zf:
        assert "org.dairiki.inkex_bh/METADATA.json" in zf.namelist()


class DummyTarget:
    def __init__(self, path: Path):
        path.mkdir(exist_ok=True)
        self.path = path

    def add_dist(
        self,
        dir_name: str | None = None,
        **metadata: str,  # name and version, mostly
    ) -> Path:
        if dir_name:
            dist_path = self.path / dir_name
        else:
            base_name = re.sub(r"(?i)[^a-z0-9]", "_", metadata.get("name", "distdir"))
            dist_paths = (self.path / f"{base_name}_{n}" for n in itertools.count(1))
            for dist_path in dist_paths:
                if not dist_path.exists():
                    break
        dist_path.mkdir(parents=True)
        Path(dist_path, "junk.txt").touch()
        if metadata:
            Path(dist_path, "METADATA.json").write_text(json.dumps(metadata))
        return dist_path

    def __fspath__(self) -> str:
        return os.fspath(self.path)


@pytest.fixture
def target(tmp_path: Path) -> DummyTarget:
    return DummyTarget(tmp_path / "target")


def test_find_installed(target: DummyTarget) -> None:
    target.add_dist("not_a_dist")
    target.add_dist(name="wrong-proj", version="1.0")
    dist_path = target.add_dist(name="test_proj", version="1.2")

    assert find_installed(target, canonicalize_name("test.proj")) == {
        Version("1.2"): dist_path,
    }


@pytest.mark.parametrize("dir_exists", [False, True])
def test_find_installed_raises(dir_exists: bool, tmp_path: Path) -> None:
    distdir = tmp_path / "distdir"
    if dir_exists:
        distdir.mkdir()
    assert find_installed(distdir, canonicalize_name("inkex-bh")) == {}


datetime_now = datetime.datetime.now(datetime.timezone.utc)


def make_asset(
    browser_download_url: str = "https://example.org/download",
    updated_at: datetime.datetime = datetime_now,
    content_type: str = "application/zip",
) -> github.ReleaseAsset:
    return github.ReleaseAsset(
        browser_download_url=browser_download_url,
        updated_at=updated_at,
        content_type=content_type,
        name="",
        label=None,
        state="uploaded",
        download_count=1,
        size=23,
        created_at=datetime_now,
    )


def make_release(
    tag_name: str, assets: Iterable[github.ReleaseAsset] = ()
) -> github.Release:
    return github.Release(
        tag_name=tag_name,
        assets=list(assets),
        name=None,
        body=None,
        html_url="",
        draft=False,
        prerelease=False,
        created_at=datetime_now,
        published_at=None,
    )


def test_find_distributions(mocker: MockerFixture) -> None:
    releases = [
        make_release("bad_tag_name"),
        make_release(
            "v1.2",
            [
                make_asset("old", datetime_now - datetime.timedelta(20)),
                make_asset("new"),
            ],
        ),
        make_release(
            "v1.5dev3+local",
            [
                make_asset("text", content_type="text/plain"),
                make_asset("zip"),
            ],
        ),
        make_release("v0.2", []),
    ]
    github = mocker.patch("barnhunt.installer.github")
    github.iter_releases.return_value = iter(releases)

    project = InkexProject("tests", "foo", "bar")

    assert find_distributions(project) == {
        Version("1.2"): "new",
        Version("1.5dev3+local"): "zip",
    }
    github.iter_releases.assert_called_once_with("foo", "bar", github_token=None)


@pytest.fixture
def target1(target: DummyTarget) -> Path:
    target.add_dist("extensions/other", name="other", version="1.0")
    target.add_dist("extensions/nondist")
    target.add_dist("symbols/old", name="bh-symbols", version="0.1rc1")
    return target.path


class ZipMaker:
    def __init__(self, zip_dir: Path):
        zip_dir.mkdir(exist_ok=True)
        self.zip_dir = zip_dir
        self._zip_files = (zip_dir / f"test{n}.zip" for n in itertools.count(1))

    def __call__(self, install_dir: str = "new", **metadata: str) -> str:
        zip_file = next(self._zip_files)
        with zipfile.ZipFile(zip_file, "w") as zf:
            zf.writestr(f"{install_dir}/METADATA.json", json.dumps(metadata))
        return urlunsplit(("file", "", str(PurePosixPath(zip_file.resolve())), "", ""))


@pytest.fixture
def zip_maker(tmp_path: Path) -> ZipMaker:
    return ZipMaker(tmp_path / "test-zips")


@pytest.mark.requiresinternet
@mayberatelimited
def test_Installer_install_from_gh(
    target: DummyTarget, caplog: pytest.LogCaptureFixture
) -> None:
    target.add_dist("extensions/other", name="other", version="1.0")
    target.add_dist("extensions/nondist")
    installer = Installer(target, github_token=os.environ.get("GITHUB_TOKEN"))
    installer.install(InkexRequirement("inkex-bh==1.0.0rc3"))
    assert {p.name for p in Path(target, "extensions").iterdir()} == {
        "other",
        "nondist",
        "org.dairiki.inkex_bh",
    }
    assert "uninstalling" not in caplog.text
    assert "installing inkex-bh==1.0.0rc3" in caplog.text


@pytest.mark.parametrize("dry_run", [False, True])
def test_Installer_install_from_file(
    dry_run: bool,
    target: DummyTarget,
    zip_maker: ZipMaker,
    caplog: pytest.LogCaptureFixture,
) -> None:
    target.add_dist("symbols/old", name="bh-symbols", version="0.1rc1")
    installer = Installer(target, dry_run=dry_run)
    download_url = zip_maker(name="bh_symbols", version="0.1a1+test")
    installer.install(InkexRequirement(f"bh-symbols @ {download_url}"))
    assert {p.name for p in Path(target, "symbols").iterdir()} == {
        "old" if dry_run else "new"
    }
    assert "uninstalling bh-symbols==0.1rc1" in caplog.text
    assert "installing bh-symbols==0.1a1+test" in caplog.text


def test_Installer_install_already_installed(
    target: DummyTarget, caplog: pytest.LogCaptureFixture
) -> None:
    target.add_dist("symbols/old", name="bh-symbols", version="0.1rc1")
    installer = Installer(target)
    installer.install(InkexRequirement("bh.symbols"))
    assert {p.name for p in Path(target, "symbols").iterdir()} == {"old"}
    assert "bh.symbols==0.1rc1 is already installed" in caplog.text


def test_Installer_install_no_distribution_found(
    target: DummyTarget, caplog: pytest.LogCaptureFixture, mocker: MockerFixture
) -> None:
    github = mocker.patch("barnhunt.installer.github")
    github.iter_releases.return_value = iter([])
    installer = Installer(target)
    with pytest.raises(NoSuchDistribution):
        installer.install(InkexRequirement("inkex-bh"))


def test_Installer_install_up_to_date(
    target: DummyTarget, caplog: pytest.LogCaptureFixture, mocker: MockerFixture
) -> None:
    target.add_dist("symbols/old", name="bh-symbols", version="0.1rc1")
    github = mocker.patch("barnhunt.installer.github")
    github.iter_releases.return_value = iter([make_release("0.1rc1", [make_asset()])])
    installer = Installer(target)
    installer.install(InkexRequirement("bh_symbols"), upgrade=True)
    assert "bh_symbols==0.1rc1 is up-to-date" in caplog.text


def test_Installer_install_multiple_installed(
    target: DummyTarget, caplog: pytest.LogCaptureFixture
) -> None:
    target.add_dist("symbols/dup1", name="bh-symbols", version="0.1rc1")
    target.add_dist("symbols/dup2", name="bh-symbols", version="0.2")
    installer = Installer(target)
    installer.install(InkexRequirement("bh.symbols"))
    assert {p.name for p in Path(target, "symbols").iterdir()} == {"dup1", "dup2"}
    assert "found multiple installed versions of bh.symbols" in caplog.text
    assert "bh.symbols==0.2 is already installed" in caplog.text


@pytest.mark.parametrize("pre_flag", [False, True])
def test_Installer_install_pre_flag(
    target: DummyTarget,
    pre_flag: bool,
    zip_maker: ZipMaker,
    caplog: pytest.LogCaptureFixture,
    mocker: MockerFixture,
) -> None:
    github = mocker.patch("barnhunt.installer.github")
    github.iter_releases.return_value = (
        make_release(version, [make_asset(zip_maker(name="inkex-bh", version=version))])
        for version in ("1.0.1", "1.1rc1")
    )

    installer = Installer(target)
    installer.install(InkexRequirement("inkex-bh"), pre_flag=pre_flag)

    assert {p.name for p in Path(target, "extensions").iterdir()} == {"new"}

    selected, unselected = ("1.1rc1", "1.0.1") if pre_flag else ("1.0.1", "1.1rc1")
    assert f"installing inkex-bh=={selected}" in caplog.text
    assert f"installing inkex-bh=={unselected}" not in caplog.text


@pytest.mark.parametrize("was_installed", [False, True])
def test_Installer_uninstall(
    was_installed: bool,
    target: DummyTarget,
    caplog: pytest.LogCaptureFixture,
) -> None:
    target.add_dist("extensions/other", name="other", version="1.0")
    target.add_dist("extensions/nondist")
    if was_installed:
        target.add_dist("extensions/old", name="inkex_bh", version="42.0")
    installer = Installer(target)
    installer.uninstall(InkexRequirement("inkex-bh"))
    assert {p.name for p in Path(target, "extensions").iterdir()} == {
        "other",
        "nondist",
    }
    if was_installed:
        assert "uninstalling inkex-bh==42.0" in caplog.text
    else:
        assert len(caplog.records) == 0



from __future__ import annotations

import logging
import os
import re
from pathlib import Path

import pytest
from click.testing import CliRunner
from lxml import etree
from pdfminer.high_level import extract_text

from barnhunt.cli import main


def test_random_seed(tmp_drawing_svg: Path, caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.INFO)
    svgfile = tmp_drawing_svg.__fspath__()

    def get_random_seed() -> int:
        tree = etree.parse(svgfile)
        root = tree.getroot()
        assert root.tag == "{http://www.w3.org/2000/svg}svg"
        value = root.attrib[
            "{http://dairiki.org/barnhunt/inkscape-extensions}random-seed"
        ]
        return int(value)

    def run_it(*args: str) -> None:
        cmd = ["random-seed"]
        cmd.extend(args)
        cmd.append(svgfile)
        runner = CliRunner()
        result = runner.invoke(main, cmd)
        assert result.exit_code == 0

    # Set seed in SVG file without pre-existing seed
    run_it()
    random_seed = get_random_seed()
    assert isinstance(random_seed, int)

    # Check that existing seed is not overwritten
    caplog.clear()
    run_it()
    assert get_random_seed() == random_seed
    assert re.search(r"\balready\b.*\bset\b", caplog.text)

    # Force overwriting of existin seed
    caplog.clear()
    run_it("--force-reseed")
    assert get_random_seed() != random_seed
    assert not re.search(r"\balready\b.*\bset\b", caplog.text)


@pytest.mark.parametrize("processes", [None, "1"])
@pytest.mark.usefixtures("inkscape_executable")  # skip if no inkscape
def test_pdfs(
    tmp_path: Path, caplog: pytest.LogCaptureFixture, processes: str | None
) -> None:
    caplog.set_level(logging.INFO)
    here = os.path.dirname(__file__)
    drawing_svg = os.path.join(here, "drawing.svg")
    cmd = ["pdfs", "-o", os.fspath(tmp_path), drawing_svg]
    if processes is not None:
        cmd[1:1] = ["-p", processes]
    runner = CliRunner()
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    expected_pdfs = {
        "novice.pdf",
        "Master_1/Blind_1.pdf",
    }
    outputs = {
        Path(dirpath).joinpath(fn)
        for dirpath, _, filenames in os.walk(tmp_path)
        for fn in filenames
    }
    assert outputs == {tmp_path.joinpath(pdf) for pdf in expected_pdfs}

    # Check that template was expanded
    assert "Novice 1" in extract_text(tmp_path / "novice.pdf", page_numbers=[0])


def test_rats() -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["rats"])
    assert result.exit_code == 0
    assert re.sub(r"[1-5]", "X", result.output) == ("X X X X X\n" * 5)


def test_coords() -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["coords", "-n", "50"])
    assert result.exit_code == 0
    lines = result.output.rstrip().split("\n")
    pairs = [list(map(int, line.split(","))) for line in lines if line]
    assert len(pairs) == 50
    for x, y in pairs:
        assert 0 <= x <= 25
        assert 0 <= y <= 30


def test_2up(tmp_path: Path, test1_pdf: Path) -> None:
    outfile = tmp_path / "output.pdf"
    runner = CliRunner()
    result = runner.invoke(
        main,
        ["2up", "-o", os.fspath(outfile), os.fspath(test1_pdf)],
    )
    assert result.exit_code == 0
    assert outfile.exists()



from __future__ import annotations

import re
import time
from contextlib import ExitStack
from itertools import cycle
from itertools import islice
from pathlib import Path
from typing import Sequence

import pikepdf
import pytest

from barnhunt.pdfutil import concat_pdfs
from barnhunt.pdfutil import iso_date
from barnhunt.pdfutil import two_up
from barnhunt.pdfutil import update_metadata


@pytest.fixture
def input_pdfs(test1_pdf: Path, test2_pdf: Path, n_pdfs: int) -> list[Path]:
    return list(islice(cycle((test1_pdf, test2_pdf)), n_pdfs))


@pytest.mark.parametrize("n_pdfs", [1, 2, 3])
def test_concat_pdfs(input_pdfs: Sequence[Path], n_pdfs: int, tmp_path: Path) -> None:
    output_fn = tmp_path / "foo/output.pdf"
    concat_pdfs(input_pdfs, output_fn)
    assert page_count(output_fn) == n_pdfs
    if n_pdfs > 1:
        assert xmp_title(output_fn) == "Test File #1"
    assert docinfo_title(output_fn) == "Test File #1"


def test_concat_pdfs_no_pdfs(tmp_path: Path) -> None:
    output_fn = tmp_path / "empty.pdf"
    with pytest.raises(ValueError):
        concat_pdfs([], output_fn)


@pytest.mark.parametrize("n_pdfs", [1, 2, 3])
def test_two_up(input_pdfs: Sequence[Path], n_pdfs: int, tmp_path: Path) -> None:
    out_path = tmp_path / "output.pdf"
    with ExitStack() as stack:
        in_files = [stack.enter_context(pdf.open("rb")) for pdf in input_pdfs]
        with out_path.open("wb") as out_file:
            two_up(in_files, out_file)
    assert page_count(out_path) == (n_pdfs + 1) // 2
    assert xmp_title(out_path) == "Test File #1"


def test_two_up_no_pages(tmp_path: Path) -> None:
    out_path = tmp_path / "empty.pdf"
    with pytest.raises(ValueError):
        with out_path.open("wb") as out_file:
            two_up([], out_file)


def test_update_metadata() -> None:
    now = time.time()
    pdf = pikepdf.Pdf.new()
    update_metadata(pdf, now_=now)
    meta = pdf.open_metadata()
    assert meta["pdf:Producer"].startswith("barnhunt ")
    assert meta["xmp:MetadataDate"] == iso_date(now)
    assert meta["xmp:ModifyDate"] == iso_date(now)


@pytest.fixture
def docinfo_test_pdf(tmp_path: Path) -> Path:
    """A test_pdf file with docinfo but no XMP metadata."""
    test_pdf = tmp_path / "docinfo-test.pdf"
    pdf = pikepdf.Pdf.new()
    pdf.docinfo.Author = "Joe"
    pdf.save(test_pdf)
    return test_pdf


@pytest.fixture
def xmp_meta_test_pdf(tmp_path: Path) -> Path:
    """A test_pdf file XMP metadata (and docinfo metadata which does not match)."""
    test_pdf = tmp_path / "docinfo-test.pdf"
    pdf = pikepdf.Pdf.new()
    with pdf.open_metadata() as meta:
        meta["dc:creator"] = ["Jane"]
    pdf.docinfo.Author = "Not Jane"
    pdf.save(test_pdf)
    return test_pdf


def test_update_metadata_from_docinfo(docinfo_test_pdf: Path) -> None:
    pdf = pikepdf.Pdf.new()
    update_metadata(pdf, pikepdf.Pdf.open(docinfo_test_pdf))
    assert pdf.docinfo.Author == "Joe"
    meta = pdf.open_metadata()
    assert meta["dc:creator"] == ["Joe"]


def test_update_metadata_from_xmp_meta(xmp_meta_test_pdf: Path) -> None:
    pdf = pikepdf.Pdf.new()
    update_metadata(pdf, pikepdf.Pdf.open(xmp_meta_test_pdf))
    assert pdf.docinfo.Author == "Jane"
    meta = pdf.open_metadata()
    assert meta["dc:creator"] == ["Jane"]


def test_iso_date() -> None:
    assert re.match(r"\d{4}(-\d\d){2}T\d\d(:\d\d){2}(\.\d+)?[-+]\d\d:\d\d", iso_date())


def page_count(pdf_fn: str | Path) -> int:
    # https://github.com/pikepdf/pikepdf/issues/452
    pdf = pikepdf.Pdf.open(pdf_fn)
    return len(pdf.pages)


def xmp_title(pdf_fn: str | Path) -> str | None:
    # https://github.com/pikepdf/pikepdf/issues/452
    pdf = pikepdf.Pdf.open(pdf_fn)
    return pdf.open_metadata().get("dc:title")


def docinfo_title(pdf_fn: str | Path) -> str | None:
    # https://github.com/pikepdf/pikepdf/issues/452
    pdf = pikepdf.Pdf.open(pdf_fn)
    title = pdf.docinfo.get("/Title")
    if title is not None:
        return str(title)
    return None



from __future__ import annotations

import ctypes
import os
import re
import sys
from pathlib import Path
from subprocess import CompletedProcess
from types import ModuleType
from types import SimpleNamespace

import pytest
from pytest_mock import MockerFixture

from barnhunt.inkscape.utils import _get_appdata
from barnhunt.inkscape.utils import get_default_user_data_directory
from barnhunt.inkscape.utils import get_user_data_directory

if sys.version_info >= (3, 8):
    from ctypes import wintypes
else:
    wintypes = ModuleType("ctypes.wintypes")
    wintypes.HWND = ctypes.c_void_p
    wintypes.HANDLE = ctypes.c_void_p
    wintypes.DWORD = ctypes.c_ulong
    wintypes.LPWSTR = ctypes.c_wchar_p
    wintypes.MAX_PATH = 260


def test_get_user_data_directory(mocker: MockerFixture) -> None:
    mocker.patch("shutil.which", return_value="/bin/inkscape")
    mocker.patch(
        "barnhunt.inkscape.utils.run",
        return_value=CompletedProcess((), 0, "cruft\n/path/to/config\n"),
    )
    assert get_user_data_directory() == Path("/path/to/config")


def test_get_user_data_directory_no_output(
    mocker: MockerFixture, caplog: pytest.LogCaptureFixture
) -> None:
    mocker.patch("shutil.which", return_value="/bin/inkscape")
    mocker.patch(
        "barnhunt.inkscape.utils.run", return_value=CompletedProcess((), 0, "")
    )
    assert get_user_data_directory() == get_default_user_data_directory()
    assert "did not produce any output" in caplog.text


def test_get_user_data_directory_no_inkscape(
    mocker: MockerFixture, caplog: pytest.LogCaptureFixture
) -> None:
    mocker.patch("shutil.which", return_value=None)
    assert get_user_data_directory() == get_default_user_data_directory()
    assert re.search(r"command .* not found", caplog.text)


@pytest.mark.skipif(sys.platform == "win32", reason="not POSIX system")
def test_get_default_user_data_directory(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delitem(os.environ, "XDG_CONFIG_HOME", raising=False)
    assert get_default_user_data_directory() == Path.home() / ".config" / "inkscape"


@pytest.mark.skipif(sys.platform == "win32", reason="not POSIX system")
def test_get_default_user_data_directory_from_env(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setitem(os.environ, "XDG_CONFIG_HOME", "/path/to/config")
    assert get_default_user_data_directory() == Path("/path/to/config/inkscape")


@pytest.fixture
def mock_windll_SHGetFolderPathW(monkeypatch: pytest.MonkeyPatch) -> None:
    # OCD to get test coverage of the win32 bits under linux
    if sys.platform == "win32":
        return
    if not os.environ.get("BARNHUNT_MOCK_WIN32"):
        pytest.skip("BARNHUNT_MOCK_WIN32 not set")

    monkeypatch.setattr("sys.platform", "win32")
    appdata = Path.home() / "AppData" / "Roaming"

    def SHGetFolderPathW(
        _hwnd: wintypes.HWND,
        csidl: ctypes.c_int,
        _hToken: wintypes.HANDLE,
        _dwFlags: wintypes.DWORD,
        pszPath: wintypes.LPWSTR,
    ) -> int:
        assert csidl.value == 26
        pszPath.value = os.fspath(appdata)
        return 0

    monkeypatch.setattr(
        "ctypes.windll",
        SimpleNamespace(shell32=SimpleNamespace(SHGetFolderPathW=SHGetFolderPathW)),
        raising=False,
    )
    monkeypatch.setitem(sys.modules, "ctypes.wintypes", wintypes)


@pytest.mark.skipif(sys.platform != "win32", reason="not Windows system")
def test_get_default_user_data_directory_from_appdata() -> None:
    default_user_data_directory = get_default_user_data_directory()
    assert default_user_data_directory.is_absolute()
    assert default_user_data_directory.name == "inkscape"


@pytest.mark.skipif(sys.platform == "win32", reason="not POSIX system")
@pytest.mark.usefixtures("mock_windll_SHGetFolderPathW")
def test_get_default_user_data_directory_from_appdata_coverage() -> None:
    default_user_data_directory = get_default_user_data_directory()
    assert default_user_data_directory.is_absolute()
    assert default_user_data_directory.name == "inkscape"


@pytest.mark.skipif(sys.platform != "win32", reason="not Windows system")
def test_get_appdata(monkeypatch: pytest.MonkeyPatch) -> None:
    appdata = _get_appdata()
    assert Path(appdata).is_absolute()



import sys

if sys.version_info < (3, 8):
    import importlib_metadata as metadata
    from typing_extensions import Literal
    from typing_extensions import Protocol
else:
    from importlib import metadata
    from typing import Literal
    from typing import Protocol


if sys.version_info < (3, 10):
    from typing_extensions import TypeGuard
else:
    from typing import TypeGuard

__all__ = ["Literal", "metadata", "Protocol", "TypeGuard"]



from __future__ import annotations

import logging
import os
import random
import sys
from collections import defaultdict
from contextlib import ExitStack
from functools import partial
from itertools import count
from multiprocessing.pool import ThreadPool
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any
from typing import BinaryIO
from typing import Callable
from typing import Iterable
from typing import Sequence
from typing import TypeVar

import click
from atomicwrites import atomic_write

from .coursemaps import Coursemap
from .coursemaps import iter_coursemaps
from .inkscape.runner import inkscape_runner
from .inkscape.utils import get_user_data_directory
from .installer import DEFAULT_REQUIREMENTS
from .installer import InkexRequirement
from .installer import Installer
from .pager import get_pager
from .pdfutil import concat_pdfs
from .pdfutil import two_up

_T = TypeVar("_T")
_U = TypeVar("_U")

log = logging.getLogger("")

POSITIVE_INT = click.IntRange(1, None)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.version_option()
def main(verbose: int) -> None:
    """Utilities for creating Barn Hunt course maps."""
    log_level = logging.WARNING
    if verbose:  # pragma: NO COVER
        log_level = logging.DEBUG if verbose > 1 else logging.INFO
    logging.basicConfig(
        level=log_level, format="(%(levelname)1.1s) [%(threadName)s] %(message)s"
    )


@main.command()
@click.argument(
    "svgfiles",
    type=click.Path(exists=True, dir_okay=False, writable=True),
    nargs=-1,
    required=True,
)
@click.option(
    "--force-reseed/--no-force-reseed",
    "-f",
    help="Force reseeding, even if a seed has been previously set.",
)
def random_seed(svgfiles: Iterable[str], force_reseed: bool) -> None:
    """Set random-seem for SVG file.

    This command sets a random random seed in the named SVG files.
    The random seed is used, e.g., when generating random rat numbers.
    Having the seed specified in the source SVG file ensures that the
    random rat numbers are reproduceable.

    By default, this will only set a random seed if one has not
    already been set.  Use the --force-reseed to override this
    behavior.
    """
    from lxml import etree

    from barnhunt.inkscape import svg

    for svgpath in svgfiles:
        tree = etree.parse(svgpath)
        if not force_reseed:
            if svg.get_random_seed(tree) is not None:
                log.info("%s: already has random-seed set, skipping", svgpath)
                continue
        random_seed = random.randrange(2**128)
        log.debug("%s: setting random seed to %d", svgpath, random_seed)
        svg.set_random_seed(tree, random_seed)
        with atomic_write(svgpath, mode="wb", overwrite=True) as f:
            tree.write(f)


def default_inkscape_command() -> str:
    # This is what inkex.command does to find Inkscape (after first
    # checking $INKSCAPE_COMMAND).
    #
    # https://gitlab.com/inkscape/extensions/-/blob/cb74374e46894030775cf947e97ca341b6ed85d8/inkex/command.py#L45
    if sys.platform == "win32":
        # prefer inkscape.exe over inkscape.com which spawns a command window
        return "inkscape.exe"
    return "inkscape"


def inkscape_command_option(**kwargs: Any) -> Callable[..., Any]:
    return click.option(
        "--inkscape-command",
        "--inkscape",
        metavar="COMMAND",
        envvar="INKSCAPE_COMMAND",  # NB: this is what inkex uses
        default=default_inkscape_command,
        help=f"""
        Name of (or path to) inkscape executable to use for exporting PDFs and for
        determining the location of the user profile directory.
        (Equivalently, you may set the $INKSCAPE_COMMAND environment variable.)
        The default is {default_inkscape_command()!r}.
        """,
        **kwargs,
    )


def default_shell_mode() -> bool:
    """Whether to use Inkscape's shell-mode by default."""
    if sys.platform == "darwin":
        return False  # ShellModeRunner current borked
    return True


@main.command()
@click.argument("svgfiles", type=click.File("rb"), nargs=-1, required=True)
@click.option(
    "--output-directory",
    "-o",
    type=click.Path(file_okay=False),
    default="pdfs",
    help="""
    Directory into which to write output PDF files.
    The default is './pdfs'.
    """,
)
@click.option(
    "--processes",
    "-p",
    metavar="N",
    type=POSITIVE_INT,
    default=os.cpu_count,
    help="""
    Number of inkscape processes to run in parallel.
    Set to one to disable parallel processing.
    The default is {os.cpu_count()} (the number of CPUs detected on this platform).
    """,
)
@inkscape_command_option()
@click.option(
    "--shell-mode-inkscape/--no-shell-mode-inkscape",
    "shell_mode",
    default=default_shell_mode,
    help="""
    Enable/disable running inkscape in shell-mode for efficiency.
    The default is enabled (except on macOS, where shell-mode currently
    seems to be broken).
    """,
)
def pdfs(
    svgfiles: Iterable[BinaryIO],
    output_directory: str,
    shell_mode: bool,
    inkscape_command: str,
    processes: int,
) -> None:
    """Export PDFs from inkscape SVG coursemaps."""
    counter = count(1)

    with ExitStack() as stack:
        tmpdir = stack.enter_context(TemporaryDirectory())
        inkscape = stack.enter_context(
            inkscape_runner(shell_mode=shell_mode, executable=inkscape_command)
        )

        def write_pdf(coursemap: Coursemap) -> tuple[Coursemap, str]:
            """Write coursemap to SVG, render to PDF in tmpdir"""
            n = next(counter)
            svg_fn = os.path.join(tmpdir, f"in{n}.svg")
            out_fn = os.path.join(tmpdir, f"out{n}.pdf")
            with open(svg_fn, "wb") as fp:
                coursemap.tree.write(fp)

            inkscape.export_pdf(svg_fn, out_fn)
            os.unlink(svg_fn)
            return coursemap, out_fn

        if processes == 1:
            map_: Callable[[Callable[[_T], _U], Iterable[_T]], Iterable[_U]] = map
        else:
            pool = stack.enter_context(ThreadPool(processes))
            map_ = pool.imap_unordered

        pages = defaultdict(list)
        for coursemap, temp_fn in map_(write_pdf, iter_coursemaps(svgfiles)):
            log.info("Rendered %s", coursemap.description)
            output_fn = os.path.join(output_directory, f"{coursemap.basename}.pdf")
            pages[output_fn].append((coursemap, temp_fn))

        for output_fn, render_info in pages.items():
            coursemaps, temp_fns = zip(*sorted(render_info))
            if log.isEnabledFor(logging.INFO):
                for coursemap in coursemaps:
                    log.info("Reading %s", coursemap.description)

            concat_pdfs(temp_fns, output_fn)
            log.warning("Wrote %d pages to %r", len(temp_fns), os.fspath(output_fn))


@main.command("rats")
@click.option(
    "-n",
    "--number-of-rows",
    type=POSITIVE_INT,
    metavar="<n>",
    help="Number of rows of rat numbers to generate.  (Default: 5).",
    default=5,
)
def rats_(number_of_rows: int) -> None:
    """Generate random rat counts.

    Prints rows of five random numbers in the range [1, 5].
    """
    for _ in range(number_of_rows):
        rats = tuple(random.randint(1, 5) for n in range(5))
        print("%d %d %d %d %d" % rats)


@main.command()
@click.option(
    "-n",
    "--number-of-rows",
    type=POSITIVE_INT,
    default=1000,
    metavar="<n>",
    help="Number of coordinates to generate. "
    "(Default: 1000 or the number of points in the grid, "
    "whichever is fewer).",
)
@click.option(
    "-g",
    "--group-size",
    type=POSITIVE_INT,
    metavar="<n>",
    help="Group output in chunks of this size. "
    "Blank lines will be printed between groups. "
    "(Default: 10).",
    default=10,
)
@click.argument(
    "dimensions",
    nargs=2,
    type=POSITIVE_INT,
    metavar="[<x-max> <y-max>]",
    envvar="BARNHUNT_DIMENSIONS",
    default=(25, 30),
)
def coords(dimensions: tuple[int, int], number_of_rows: int, group_size: int) -> None:
    """Generate random coordinates.

    Generates random coordinates.  The coordinates will range between (0, 0)
    and the (<x-max>, <y-max>).  Duplicates will be eliminated.

    The course dimensions may also be specified via
    BARNHUNT_DIMENSIONS environment variable.  E.g.

        export BARNHUNT_DIMENSIONS="25 30"

    """
    x_max, y_max = dimensions

    dim_x = dimensions[0] + 1
    dim_y = dimensions[1] + 1
    n_pts = dim_x * dim_y
    number_of_rows = min(number_of_rows, n_pts)

    def coord(pt: int) -> tuple[int, int]:
        y, x = divmod(pt, dim_x)
        return x, y

    pager = get_pager(group_size)
    pager(
        [
            "{0[0]:3d},{0[1]:3d}".format(coord(pt))
            for pt in random.sample(range(n_pts), number_of_rows)
        ]
    )


def default_2up_output_file() -> Path:
    """Compute default output filename."""
    ctx = click.get_current_context()
    input_paths = {Path(infp.name) for infp in ctx.params.get("pdffiles", ())}
    if len(input_paths) != 1:
        raise click.UsageError(
            "Can not deduce default output filename when multiple input "
            "files are specified.",
            ctx=ctx,
        )
    input_path = input_paths.pop()
    output_path = input_path.with_name(input_path.stem + "-2up" + input_path.suffix)
    click.echo(f"Writing output to {output_path!s}")
    return output_path


@main.command(name="2up")
@click.argument("pdffiles", type=click.File("rb"), nargs=-1, required=True)
@click.option(
    "-o",
    "--output-file",
    type=click.File("wb", atomic=True),
    default=default_2up_output_file,
    help="Output file name. " "(Default input filename with '-2up' appended to stem.)",
)
def pdf_2up(pdffiles: Iterable[BinaryIO], output_file: BinaryIO) -> None:
    """Format PDF(s) for 2-up printing.

    Pages printed "pre-shuffled".  The first half of the input pages
    will be printed on the top half of the output pages, and the
    second half on the lower part of the output pages.  This way, the
    resulting stack out output can be cut in half, and the pages will
    be in proper order without having to shuffle them.

    """
    two_up(pdffiles, output_file)


class InkexRequirementType(click.ParamType):
    name = "requirement"

    def __init__(self, allow_specifiers: bool = True):
        self.allow_specifiers = allow_specifiers

    def convert(
        self,
        value: str | InkexRequirement | None,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> InkexRequirement:
        if not isinstance(value, InkexRequirement):
            try:
                value = InkexRequirement(value or "")
            except ValueError as exc:
                self.fail(str(exc), param, ctx)
        if not self.allow_specifiers:
            if value.specifier:
                self.fail("specifiers not allowed in requirement")
            if value.url:
                self.fail("direct reference (URL) not allowed in requirement")
        return value


target_option = click.option(
    "--target",
    type=click.Path(exists=True, file_okay=False, writable=True, path_type=Path),
    envvar="INKSCAPE_PROFILE_DIR",
    help="""
    Where to install distributions.
    Defaults to inkscape’s user data directory, e.g. $XDG_CONFIG_HOME/inkscape, or
    %APPDATA%\\inkscape on Windows.  This may also be set by setting the
    $INKSCAPE_PROFILE_DIR environment variable.""",
)


def set_default_target(
    ctx: click.Context, param: click.Parameter, inkscape_command: str
) -> str:
    if ctx.default_map is None:
        ctx.default_map = {}
    ctx.default_map["target"] = partial(get_user_data_directory, inkscape_command)
    return inkscape_command


@main.command()
@click.argument(
    "requirements",
    type=InkexRequirementType(),
    nargs=-1,
)
@click.option(
    "--upgrade/--no-upgrade", "-U", help="Upgrade to the newest available versions."
)
@click.option("--pre/--no-pre", help="Include pre-release and development versions.")
@click.option("-n", "--dry-run/--no-dry-run", help="Just show what would be done.")
@target_option
@inkscape_command_option(
    is_eager=True,
    expose_value=False,
    callback=set_default_target,
)
@click.option(
    "--github-token",
    envvar="GITHUB_TOKEN",
    help=""" Token to use for authentication to GitHub’s REST API.
    This is not typically necessary, but if you are seeing rate-limit errors, using a
    GitHub “Personal Access Token” here (it does not need to be granted access to any
    scopes) greatly increases the rate-limit thresholds.  The token may also be passed
    in the $GITHUB_TOKEN environment variable. """,
)
def install(
    target: Path,
    upgrade: bool,
    pre: bool,
    dry_run: bool,
    github_token: str | None,
    requirements: Sequence[InkexRequirement],
) -> None:
    """Install extensions and/or symbol sets."""
    installer = Installer(target, dry_run=dry_run, github_token=github_token)
    log.warning("installing to '%s'", target)
    for requirement in requirements or DEFAULT_REQUIREMENTS:
        installer.install(requirement, pre_flag=pre, upgrade=upgrade)


@main.command()
@click.argument(
    "requirements",
    type=InkexRequirementType(allow_specifiers=False),
    nargs=-1,
)
@click.option("-n", "--dry-run/--no-dry-run", help="Just show what would be done.")
@target_option
@inkscape_command_option(
    is_eager=True,
    expose_value=False,
    callback=set_default_target,
)
def uninstall(
    target: Path,
    dry_run: bool,
    requirements: Sequence[InkexRequirement],
) -> None:
    """Uninstall extensions and/or symbol sets."""
    installer = Installer(target, dry_run=dry_run)
    for requirement in requirements or DEFAULT_REQUIREMENTS:
        installer.uninstall(requirement)



from __future__ import annotations

import hashlib
import logging
import os
import random
import re
from functools import partial
from typing import Any
from typing import BinaryIO
from typing import Dict
from typing import Iterator
from typing import Sequence

import jinja2

from .inkscape import svg
from .layerinfo import LayerFlags
from .layerinfo import LayerInfoParser
from .layerinfo import parse_flagged_layer_info

TemplateContext = Dict[str, Any]


log = logging.getLogger()


class LayerAdapter:
    """Adapt an Inkscape SVG layer element for ease of use in templates"""

    def __init__(
        self,
        elem: svg.LayerElement,
        layer_info_parser: LayerInfoParser = parse_flagged_layer_info,
    ):
        self.elem = elem
        self._parse_layer_info = layer_info_parser
        self._info = layer_info_parser(elem)

    @property
    def id(self) -> str | None:
        return self.elem.get("id")

    @property
    def label(self) -> str:
        return self._info.label

    @property
    def output_basenames(self) -> Sequence[str] | None:
        return self._info.output_basenames or None

    @property
    def is_overlay(self) -> bool:
        return bool(self._info.flags & LayerFlags.OVERLAY)

    @property
    def parent(self) -> LayerAdapter | None:
        parent = svg.parent_layer(self.elem)
        if parent is None:
            return None
        return LayerAdapter(parent, self._parse_layer_info)

    @property
    def lineage(self) -> Iterator[LayerAdapter]:
        ancestor: LayerAdapter | None = self
        while ancestor:
            yield ancestor
            ancestor = ancestor.parent

    @property
    def overlay(self) -> LayerAdapter | None:
        for layer in self.lineage:
            if layer.is_overlay:
                return layer
        return None

    def __hash__(self) -> int:
        assert self.id
        return _hash_string(self.id)

    def __eq__(self, other: Any) -> bool:
        return (
            type(other) is type(self)
            and other.elem == self.elem
            and other._parse_layer_info is self._parse_layer_info
        )

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id}>"

    def __str__(self) -> str:
        return self.label


def _hash_string(s: str) -> int:
    """A deterministic string hashing function.

    (Python's hash() can depend on the setting of PYTHONHASHSEED.)
    """
    bytes_ = s.encode("utf8")
    return hash(int(hashlib.sha1(bytes_).hexdigest(), 16))


def get_element_context(
    elem: svg.Element, layer_info_parser: LayerInfoParser = parse_flagged_layer_info
) -> TemplateContext:
    for layer in svg.lineage(elem):
        if svg.is_layer(layer):
            break
    else:
        return {}

    layer_adapter = LayerAdapter(layer, layer_info_parser)
    overlays: list[LayerAdapter] = []
    for ancestor in layer_adapter.lineage:
        if ancestor.is_overlay:
            overlays.insert(0, ancestor)

    context = {
        "layer": layer_adapter,
        "overlays": tuple(overlays),
    }
    if overlays:
        # Course is the outermost containing overlay
        context["course"] = overlays[0]
    if len(overlays) > 1:
        # Overlay is the innermost containing overlay, but only if is
        # distinct from course.
        context["overlay"] = overlays[-1]
    return context


class FileAdapter:
    """Adapt a file object for ease of use in templates"""

    def __init__(self, fp: BinaryIO):
        self._fp = fp

    @property
    def name(self) -> str:
        return self._fp.name

    @property
    def basename(self) -> str:
        return os.path.basename(self.name)

    @property
    def stat(self) -> os.stat_result:
        fd = self._fp.fileno()
        return os.fstat(fd)

    def __hash__(self) -> int:
        st = self.stat
        return hash((st.st_dev, st.st_ino))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.name!r}>"

    def __str__(self) -> str:
        return self.name


global_rng = random.Random()


def default_random_seed(context: jinja2.runtime.Context) -> int:
    random_seed = context.get("random_seed")
    layer = context.get("layer")
    return hash((random_seed, layer))


@jinja2.pass_context
def random_rats(
    context: jinja2.runtime.Context,
    n: int = 5,
    min: int = 1,
    max: int = 5,
    seed: int | None = None,
    skip: int = 0,
) -> tuple[int, ...]:
    """Generate random rat numbers.

    Returns a tuple of ``n`` random integers in the range [``min``,
    ``max``].

    By default, on a given Inkscape layer in a given file, this
    function will attempt to return the same result each time it is
    called.  See the description of the ``seed`` parameter, below, for
    more details.

    The hash of the value specified for ``seed`` is used to seed
    the random number generator used to generate the rat counts.
    The same counts will be generated for the same seed.

    If no ``seed`` is specified, the value of ``hash(random_seed,
    layer)``, where ``random_seed``, and ``layer`` are obtained from
    the current template context (with any missing values replaced
    with ``None``), will be used for the seed.  The ``random_seed``,
    if not specified in the input SVG file is constructed from the
    device and inode number of the input SVG file.  The hash of
    ``layer`` depends on the XML id of the layer.

    As a special case, passing ``seed``=``False`` will cause a global
    random number generator to be used, and it will not be reseeded.
    As a result subsequent calls with ``seed`` set to ``False`` will
    generate different rat counts.

    A positive value is passed for ``skip`` will cause that many
    random number to be generated and discarded before picking the
    ``n`` random numbers which are finally used.  This can be used to
    generate independentt sets of random numbers all based on the same
    seed.

    """
    if seed is None:
        seed = default_random_seed(context)
    rng = random.Random(seed) if seed is not False else global_rng

    rand = partial(rng.randint, min, max)
    for _ in range(skip):
        rand()
    return tuple(rand() for _ in range(n))


def safepath(path_comp: Any) -> str:
    """A jinja filter to replace shell-unfriendly characters with underscore."""
    return re.sub(r"[\000-\040/\\\177\s]", "_", str(path_comp), flags=re.UNICODE)


GLOBALS = {
    "rats": random_rats,
}

FILTERS = {
    "safepath": safepath,
}


def make_jinja2_environment(
    undefined: type[jinja2.Undefined] = jinja2.Undefined,
) -> jinja2.Environment:
    env = jinja2.Environment(autoescape=False, undefined=undefined)
    env.globals.update(GLOBALS)
    env.filters.update(FILTERS)
    return env


default_env = make_jinja2_environment()
strict_env = make_jinja2_environment(undefined=jinja2.StrictUndefined)


def render_template(
    tmpl_string: str, context: TemplateContext, strict_undefined: bool = False
) -> str:
    """Render string template."""
    env = strict_env if strict_undefined else default_env
    tmpl = env.from_string(tmpl_string)
    return tmpl.render(context)


def is_string_literal(tmpl_string: str) -> bool:
    """Is ``tmpl_string`` a simple string literal.

    Returns ``False`` if ``tmpl_string`` contains any Jinja2
    expressions or statements.
    """
    from jinja2 import nodes

    ast = default_env.parse(tmpl_string)
    assert isinstance(ast, nodes.Template)
    if len(ast.body) != 1 or not isinstance(ast.body[0], nodes.Output):
        return False
    output = ast.body[0]
    return len(output.nodes) == 1 and isinstance(output.nodes[0], nodes.TemplateData)



from __future__ import annotations

import os

import pytest

from barnhunt.installer.github import iter_releases

from ratelimit import mayberatelimited  # noreorder (test library)


@pytest.mark.requiresinternet
@mayberatelimited
def test_iter_releases() -> None:
    github_token = os.environ.get("GITHUB_TOKEN")

    # Pass per_page=1 to test page traversal
    for n, release in enumerate(
        iter_releases("actions", "setup-python", per_page=1, github_token=github_token)
    ):
        assert release.html_url.startswith("https://github.com/actions/setup-python/")
        if n:
            break
    assert n == 1



import subprocess
import sys

import barnhunt.__main__  # noqa: F401

# import for coverage


def test_execute_package() -> None:
    proc = subprocess.run(
        [sys.executable, "-m", "barnhunt", "--help"],
        stdout=subprocess.PIPE,
        text=True,
        check=True,
    )
    assert "export pdfs from inkscape" in proc.stdout.lower()



import io
import re
from types import TracebackType
from typing import AnyStr
from typing import Awaitable
from typing import Callable
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import Literal
from typing import overload
from typing import TextIO
from typing import Type
from typing import TypeVar

from _typeshed import Incomplete

from .exceptions import EOF
from .exceptions import TIMEOUT

Pattern = AnyStr | re.Pattern[AnyStr] | Type[EOF] | Type[TIMEOUT]
CompiledPattern = re.Pattern[AnyStr] | Type[EOF] | Type[TIMEOUT]
ExactPattern = AnyStr | Type[EOF] | Type[TIMEOUT]

_BufferType = TypeVar("_BufferType", io.StringIO, io.BytesIO)

class SpawnBase(Generic[AnyStr, _BufferType]):
    encoding: str | None
    pid: int | None
    flag_eof: bool
    stdin: Incomplete
    stdout: Incomplete
    stderr: Incomplete
    searcher: Incomplete
    ignorecase: bool
    before: AnyStr
    after: AnyStr
    match: re.Match[AnyStr]
    match_index: int | None
    terminated: bool
    exitstatus: int | None
    signalstatus: int | None
    status: int | None
    child_fd: int
    timeout: float
    delimiter: Incomplete
    logfile: TextIO | None
    logfile_read: TextIO | None
    logfile_send: TextIO | None
    maxread: int
    searchwindowsize: int | None
    delaybeforesend: float
    delayafterclose: float
    delayafterterminate: float
    delayafterread: float
    softspace: bool
    name: str
    closed: bool
    codec_errors: str | None
    string_type: Type[AnyStr]
    buffer_type: Type[_BufferType]
    crlf: AnyStr
    allowed_string_types: tuple[Type[bytes] | Type[str], ...]
    linesep: AnyStr
    write_to_stdout: Callable[[AnyStr], int]
    async_pw_transport: Incomplete

    @overload
    def __init__(
        self: "SpawnBase[str, io.StringIO]",
        timeout: float = ...,
        maxread: int = ...,
        searchwindowsize: int | None = ...,
        logfile: TextIO | None = ...,
        encoding: str = ...,
        codec_errors: str = ...,
    ): ...
    @overload
    def __init__(
        self: "SpawnBase[bytes, io.BytesIO]",
        timeout: float = ...,
        maxread: int = ...,
        searchwindowsize: int | None = ...,
        logfile: TextIO | None = ...,
        encoding: None = ...,
        codec_errors: str | None = ...,
    ): ...

    buffer: AnyStr

    # The default values for `size` and `timeout` are incompatible
    # with PopenSpawn.read_nonblocking which does not provide
    # defaults. We'll just leave it as virtual
    # def read_nonblocking(
    #     self, size: int = ..., timeout: float | None = ...
    # ) -> AnyStr: ...

    def compile_pattern_list(
        self, patterns: Iterable[Pattern[AnyStr]] | Pattern[AnyStr]
    ) -> list[CompiledPattern[AnyStr]]: ...
    @overload
    def expect(
        self,
        pattern: Pattern[AnyStr] | Iterable[Pattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
    ) -> int: ...
    @overload
    def expect(
        self,
        pattern: Pattern[AnyStr] | Iterable[Pattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
        *,
        async_: Literal[False],
    ) -> int: ...
    @overload
    def expect(
        self,
        pattern: Pattern[AnyStr] | Iterable[Pattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
        *,
        async_: Literal[True],
    ) -> Awaitable[int]: ...
    @overload
    def expect_list(
        self,
        pattern_list: list[CompiledPattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
    ) -> int: ...
    @overload
    def expect_list(
        self,
        pattern_list: list[CompiledPattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
        *,
        async_: Literal[False],
    ) -> int: ...
    @overload
    def expect_list(
        self,
        pattern_list: list[CompiledPattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
        *,
        async_: Literal[True],
    ) -> Awaitable[int]: ...
    @overload
    def expect_exact(
        self,
        pattern_list: Iterable[ExactPattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
    ) -> int: ...
    @overload
    def expect_exact(
        self,
        pattern_list: Iterable[ExactPattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
        *,
        async_: Literal[False],
    ) -> int: ...
    @overload
    def expect_exact(
        self,
        pattern_list: Iterable[ExactPattern[AnyStr]],
        timeout: float = ...,
        searchwindowsize: int = ...,
        *,
        async_: Literal[True],
    ) -> Awaitable[int]: ...
    def expect_loop(
        self, searcher: Incomplete, timeout: float = ..., searchwindowsize: int = ...
    ) -> int: ...
    def read(self, size: int = ...) -> AnyStr: ...
    def readline(self, size: int = ...) -> AnyStr: ...
    def __iter__(self) -> Iterator[AnyStr]: ...
    def readlines(self, sizehint: int = ...) -> list[AnyStr]: ...
    def fileno(self) -> int: ...
    def flush(self) -> None: ...
    def isatty(self) -> bool: ...

    _Self = TypeVar("_Self", bound="SpawnBase[AnyStr, _BufferType]")
    def __enter__(self: _Self) -> _Self: ...
    def __exit__(
        self,
        etype: Type[BaseException] | None,
        evalue: Exception | None,
        tb: TracebackType | None,
    ) -> None: ...



"""Support for paging output
"""
import enum
import sys
from typing import Dict
from typing import Sequence

import click

from ._compat import Protocol


class Pager(Protocol):
    def __call__(self, lines: Sequence[str]) -> None:
        ...


def get_pager(group_size: int) -> Pager:
    if sys.stdout.isatty() and sys.stdin.isatty():
        return TTYPager(group_size)
    return Grouper(group_size)


class Grouper:
    """Simple "pager" which just spews all lines to stdout, with blank
    lines every ``group_size`` lines.

    """

    def __init__(self, group_size: int):
        self.group_size = group_size

    def __call__(self, lines: Sequence[str]) -> None:
        group_size = self.group_size
        for i in range(0, len(lines), group_size):
            for line in lines[i : i + group_size]:
                print(line)
            print("")


class TTYPager:
    """Fancy terminal pager.

    Displays ``group_size`` lines per page.
    Highlights the current line.
    Allows for scrolling forward as well as backward.

    """

    def __init__(self, group_size: int):
        self.group_size = group_size

    def __call__(self, lines: Sequence[str]) -> None:
        group_size = self.group_size
        increments = {
            Command.DOWN: 1,
            Command.UP: -1,
            Command.PAGE_DOWN: group_size,
            Command.PAGE_UP: -group_size,
            Command.REDRAW: 0,
        }
        n = 0
        while lines:
            click.clear()
            beg = n - (n % group_size)
            group = lines[beg : beg + group_size]
            for i, line in enumerate(group, beg):
                hilight = i == n
                click.secho(line, bold=hilight, bg="cyan" if hilight else None)
            click.echo()
            click.echo(f"Viewing {beg + 1}–{i + 1}. [jkq] ", nl=False)

            cmd = self._get_cmd()
            if cmd == Command.QUIT:
                break
            next_n = n + increments[cmd]
            if 0 < next_n < len(lines):
                n = next_n
            else:
                click.echo("\a", nl=False)
        click.echo()

    def _get_cmd(self) -> "Command":
        while True:
            key = click.getchar()
            if key == "\x1b":
                key += click.getchar()
            try:
                return Command.lookup(key)
            except KeyError:
                click.echo("\a", nl=False)


def CTL(c: str) -> str:
    return chr(ord(c) & 0x1F)


def ESC(c: str) -> str:
    return "\x1b" + c


def ALT(c: str) -> str:
    return chr(0x100 | ord(c))


def ANSI_CSI(s: str) -> str:
    return "\x1b[" + s


K_Prior = ANSI_CSI("5~")  # Page_Up
K_Next = ANSI_CSI("6~")  # Page_Down
K_Up = ANSI_CSI("A")  # Up arrow
K_Down = ANSI_CSI("B")  # Up arrow
K_Return = "\r"
K_Escape = "\x1b"


_Command_lookup: Dict[str, "Command"] = {}


class Command(enum.Enum):
    """Commands recognized by TTYPager.

    The value of each enum is a tuple specifying which keys or key combinations
    can be used to activate the given command.

    """

    PAGE_DOWN = (
        " ",
        "v",
        CTL("v"),
        "f",
        CTL("f"),
        K_Next,
        # 'z', ESC(' '), ALT(' '),
        # 'd', CTL('d'),
    )
    PAGE_UP = (
        "b",
        CTL("b"),
        ESC("v"),
        ALT("v"),
        K_Prior,
        # 'w',
        # 'u', CTL('u'),
    )
    UP = (
        "k",
        CTL("k"),
        "y",
        CTL("y"),
        CTL("p"),
        K_Up,
    )
    DOWN = (
        "j",
        CTL("j"),
        "e",
        CTL("e"),
        K_Return,
        K_Down,
    )
    REDRAW = (
        "r",
        CTL("r"),
        CTL("l"),
        "R",
    )
    QUIT = ("q",)

    def __init__(self, *keys: str):
        _lookup = getattr(self.__class__, "_lookup", None)
        if _lookup is None:
            self.__class__._lookup = _lookup = {}  # type: ignore[attr-defined]

        for key in keys:
            assert key not in _lookup, f"redefinition of key {key!r}"
            _lookup[key] = self

    @classmethod
    def lookup(cls, key: str) -> "Command":
        """Look up comand for ``key``."""
        _lookup: Dict[str, "Command"] = cls._lookup  # type: ignore[attr-defined]
        return _lookup[key]



from __future__ import annotations

from typing import Collection
from typing import Sequence

import pytest

from barnhunt import layerinfo
from barnhunt.inkscape import svg
from barnhunt.layerinfo import dwim_layer_info
from barnhunt.layerinfo import LayerFlags
from barnhunt.layerinfo import parse_flagged_layer_info
from barnhunt.layerinfo import parse_obs_layer_info
from testlib import svg_maker


@pytest.mark.parametrize(
    "s, flags",
    [
        ("", LayerFlags(0)),
        ("h", LayerFlags.HIDDEN),
        ("o", LayerFlags.OVERLAY),
        ("ho", LayerFlags.OVERLAY | LayerFlags.HIDDEN),
        ("hoo", LayerFlags.OVERLAY | LayerFlags.HIDDEN),
        ("oh", LayerFlags.OVERLAY | LayerFlags.HIDDEN),
    ],
)
def test_layerflags_parse(s: str, flags: LayerFlags) -> None:
    assert LayerFlags.parse(s) == flags


def test_layerflags_parse_warns(caplog: pytest.LogCaptureFixture) -> None:
    assert LayerFlags.parse("hx") == LayerFlags.HIDDEN
    assert "unknown character" in caplog.text


@pytest.mark.parametrize(
    "s, flags",
    [
        ("", LayerFlags(0)),
        ("h", LayerFlags.HIDDEN),
        ("o", LayerFlags.OVERLAY),
        ("ho", LayerFlags.OVERLAY | LayerFlags.HIDDEN),
    ],
)
def test_layerflags_str(s: str, flags: LayerFlags) -> None:
    assert set(str(flags)) == set(s)


@pytest.fixture
def tree1() -> svg.ElementTree:
    layer = svg_maker.layer
    return svg_maker.tree(
        layer("Cruft"),
        layer("Ring", visible=False),
        layer(
            "T1 Master",
            children=[
                layer(
                    "Overlays",
                    children=[
                        layer("Blind 1"),
                        layer("Build Notes"),
                    ],
                )
            ],
        ),
        layer("T1 Novice"),
    )


@pytest.mark.parametrize(
    "predicate_name, expected_ids",
    [
        (
            "obs_is_ring",
            [
                "ring",
            ],
        ),
        (
            "obs_is_course",
            [
                "t1novice",
                "t1master",
            ],
        ),
        (
            "obs_is_cruft",
            [
                "cruft",
            ],
        ),
        (
            "obs_is_overlay",
            [
                "blind1",
                "buildnotes",
            ],
        ),
    ],
)
def test_predicate(
    predicate_name: str, tree1: svg.ElementTree, expected_ids: Collection[str]
) -> None:
    predicate = getattr(layerinfo, predicate_name)

    def match(elem: svg.Element) -> bool:
        return svg.is_layer(elem) and predicate(elem)

    print(list(elem.get("id") for elem in svg.walk_layers(tree1.getroot())))
    matches = {elem.get("id") for elem in tree1.iter() if match(elem)}
    assert matches == set(expected_ids)


@pytest.mark.parametrize(
    "layer_label, label, flags, output_basenames, exclude_from, include_in",
    [
        ("[h] Hidden", "Hidden", LayerFlags.HIDDEN, [], set(), set()),
        ("[o] An Overlay", "An Overlay", LayerFlags.OVERLAY, [], set(), set()),
        ("Plain Jane", "Plain Jane", LayerFlags(0), [], set(), set()),
        (
            "[o|foo] Another Overlay",
            "Another Overlay",
            LayerFlags.OVERLAY,
            ["foo"],
            set(),
            set(),
        ),
        ("[!base]Second Layer", "Second Layer", LayerFlags(0), [], {"base"}, set()),
        ("[=base]Base desc", "Base desc", LayerFlags(0), [], set(), {"base"}),
    ],
)
def test_FlaggedLayerInfo(
    layer_label: str,
    label: str,
    flags: LayerFlags,
    output_basenames: Sequence[str],
    exclude_from: Collection[str],
    include_in: Collection[str],
) -> None:
    info = parse_flagged_layer_info(svg_maker.layer(layer_label))
    # FIXME: assert info.elem is elem
    assert info.label == label
    assert info.flags is flags
    assert info.output_basenames == output_basenames
    assert info.exclude_from == exclude_from
    assert info.include_in == include_in


class Test_obs_layer_info:
    @pytest.mark.parametrize(
        "label, flags",
        [
            ("Prototypes", LayerFlags.HIDDEN),
            ("Master 1", LayerFlags.OVERLAY),
            ("Test Ring", LayerFlags(0)),
        ],
    )
    def test_init(self, label: str, flags: LayerFlags) -> None:
        info = parse_obs_layer_info(svg_maker.layer(label))
        assert info.label == label
        assert info.flags is flags
        assert len(info.output_basenames) == 0
        assert len(info.exclude_from) == 0
        assert len(info.include_in) == 0

    def test_overlay(self) -> None:
        layer = svg_maker.layer
        overlay = layer("Test ovl")
        not_overlay = layer("Test not ovl")
        svg_maker.tree(
            layer(
                "Master 2",
                children=[
                    layer("Overlays", children=[overlay]),
                    layer("Stuff", children=[not_overlay]),
                ],
            )
        )

        info = parse_obs_layer_info(overlay)
        assert info.flags is LayerFlags.OVERLAY

        info = parse_obs_layer_info(not_overlay)
        assert info.flags is LayerFlags(0)


class Test_dwim_layer_info:
    @pytest.fixture
    def dummy_tree(self, leaf_label_1: str) -> svg.ElementTree:
        layer = svg_maker.layer
        return svg_maker.tree(
            layer("Layer", children=[layer(leaf_label_1)]),
        )

    @pytest.mark.parametrize("leaf_label_1", ["[h] Hidden"])
    def test_flagged_info(self, dummy_tree: svg.ElementTree) -> None:
        assert dwim_layer_info(dummy_tree) is parse_flagged_layer_info

    @pytest.mark.parametrize("leaf_label_1", ["Not Flagged"])
    def test_obs_info(self, dummy_tree: svg.ElementTree) -> None:
        assert dwim_layer_info(dummy_tree) is parse_obs_layer_info



from functools import wraps
from typing import Any
from typing import Callable

import pytest
from requests import HTTPError


class MaybeRateLimited:
    def __init__(self) -> None:
        self.ratelimited = False

    def __call__(self, test: Callable[..., None]) -> Callable[..., None]:
        @wraps(test)
        def wrapper(*args: Any, **kwargs: Any) -> None:
            if self.ratelimited:
                pytest.xfail("GitHub API rate limit exceeded")
            try:
                test(*args, **kwargs)
            except HTTPError as exc:
                if exc.response is not None:
                    if exc.response.headers.get("x-ratelimit-remaining") == "0":
                        self.ratelimited = True
                        pytest.xfail("GitHub API rate limit exceeded")
                raise

        return wrapper


mayberatelimited = MaybeRateLimited()



class ExceptionPexpect(Exception):
    value: str
    def __init__(self, value: str) -> None: ...
    def get_trace(self) -> str: ...

class EOF(ExceptionPexpect): ...
class TIMEOUT(ExceptionPexpect): ...



import os
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Iterator

import click
import pytest
from click.testing import CliRunner
from pytest_mock import MockerFixture

from barnhunt.cli import default_2up_output_file
from barnhunt.cli import default_inkscape_command
from barnhunt.cli import default_shell_mode
from barnhunt.cli import InkexRequirementType
from barnhunt.cli import main
from barnhunt.cli import pdf_2up
from barnhunt.installer import DEFAULT_REQUIREMENTS
from barnhunt.installer import InkexRequirement


@pytest.mark.parametrize(
    "platform, expect",
    [
        ("linux", "inkscape"),
        ("win32", "inkscape.exe"),
    ],
)
def test_default_inkscape_command(
    platform: str, expect: str, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.delitem(os.environ, "INKSCAPE_COMMAND", raising=False)
    monkeypatch.setattr("sys.platform", platform)
    assert default_inkscape_command() == expect


@pytest.mark.parametrize(
    "platform, expect",
    [
        ("linux", True),
        ("darwin", False),
    ],
)
def test_default_shell_mode(
    platform: str, expect: bool, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr("sys.platform", platform)
    assert default_shell_mode() is expect


class Test_default_2up_output_file:
    @pytest.fixture
    def ctx(self) -> Iterator[click.Context]:
        with click.Context(pdf_2up) as ctx:
            yield ctx

    @pytest.fixture
    def add_input_file(
        self, ctx: click.Context, tmp_path: Path
    ) -> Callable[[str], Path]:
        def add_input_file(filename: str) -> Path:
            params = ctx.params
            param_name = "pdffiles"
            dummy_path = tmp_path / filename
            dummy_path.touch()
            fp = click.File("rb")(os.fspath(dummy_path))
            params[param_name] = params.get(param_name, ()) + (fp,)
            return dummy_path

        return add_input_file

    def test_default_output_filename(
        self,
        ctx: click.Context,
        add_input_file: Callable[[str], Path],
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        input_pdf = add_input_file("input.pdf")
        output_filename = default_2up_output_file()
        assert output_filename == input_pdf.with_name("input-2up.pdf")
        assert "Writing output to " in capsys.readouterr().out

    def test_raises_error_when_multiple_inputs(
        self, ctx: click.Context, add_input_file: Callable[[str], Path]
    ) -> None:
        add_input_file("input1.pdf")
        add_input_file("input2.pdf")
        with pytest.raises(click.UsageError, match="multiple input files"):
            default_2up_output_file()


def test_main(capsys: pytest.CaptureFixture[str]) -> None:
    with pytest.raises(SystemExit):
        main(["--help"])
    std = capsys.readouterr()
    assert "export pdfs from inkscape" in std.out.lower()


class Test_InkexRequirementType:
    CF = Callable[[Any], InkexRequirement]

    @pytest.fixture
    def allow_specifiers(self) -> bool:
        return True

    @pytest.fixture
    def convert(self, allow_specifiers: bool) -> CF:
        param_type = InkexRequirementType(allow_specifiers=allow_specifiers)

        def convert(value: Any) -> InkexRequirement:
            return param_type.convert(value, None, None)

        return convert

    def test_from_requirement(self, convert: CF) -> None:
        req = DEFAULT_REQUIREMENTS[0]
        assert convert(req) is req

    def test_from_string(self, convert: CF) -> None:
        req = convert("inkex_bh")
        assert req.name == "inkex_bh"

    def test_unknown_name(self, convert: CF) -> None:
        with pytest.raises(click.BadParameter):
            convert("unknown")

    def test_specifier(self, convert: CF) -> None:
        assert str(convert("inkex.bh<1.0.0").specifier) == "<1.0.0"

    def test_direct_reference(self, convert: CF) -> None:
        assert convert("inkex.bh @ file:///tmp/test.zip").url == "file:///tmp/test.zip"

    @pytest.mark.parametrize("allow_specifiers", [False])
    def test_allows_specifier(self, convert: CF) -> None:
        with pytest.raises(click.BadParameter):
            convert("inkex.bh==1.0.0")

    @pytest.mark.parametrize("allow_specifiers", [False])
    def test_allows_direct_reference(self, convert: CF) -> None:
        with pytest.raises(click.BadParameter):
            convert("inkex.bh @ file:///tmp/test.zip")


def test_install_target_from_inkscape(mocker: MockerFixture, tmp_path: Path) -> None:
    mocker.patch.dict("os.environ", GITHUB_TOKEN="token")
    target = tmp_path
    get_user_data_directory = mocker.patch(
        "barnhunt.cli.get_user_data_directory", return_value=target
    )
    Installer = mocker.patch("barnhunt.cli.Installer")
    runner = CliRunner()
    result = runner.invoke(main, ("install", "--inkscape", "myinkscape"))
    assert result.exit_code == 0
    get_user_data_directory.assert_called_once_with("myinkscape")
    Installer.assert_called_once_with(target, dry_run=False, github_token="token")


def test_uninstall(mocker: MockerFixture, tmp_path: Path) -> None:
    Installer = mocker.patch("barnhunt.cli.Installer")
    runner = CliRunner()
    result = runner.invoke(main, ("uninstall", "--target", os.fspath(tmp_path)))
    assert result.exit_code == 0
    assert Installer.mock_calls == [
        mocker.call(tmp_path, dry_run=False),
        *(mocker.call().uninstall(req) for req in DEFAULT_REQUIREMENTS),
    ]
